
# this is a ganga grid simulator driver script
# usage:
# env GANGA_GRID_SIMULATOR=1 ganga -o[LCG]GLITE_ENABLE=True
# -oGLITE_SETUP=/dev/null -o[PollThread]autostart=True simulator.py

# when the simulator is enabled it will produce data files in the current working directory
# these files may be further processed with the simulator-analyze.py
# script to extract timing data

# recommended way is to have the driver + simulation parameters + results all in one directory e.g.
#
# mkdir simu
# cd simu
# cp path/to/simulation.py .
# run simulation.py as described above

from Ganga.GPIDev.Lib.Job.Job import Job
from Ganga.Lib.LCG import LCG
from Ganga.Lib.Splitters import GenericSplitter
from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

config = config['GridSimulator']
config['submit_time'] = '0.2'
config['submit_failure_rate'] = 0.0
config['cancel_time'] = 'random.uniform(0,1)'
config['cancel_failure_rate'] = 0.0
config['single_status_time'] = 0.0  # * number of subjobs
config['master_status_time'] = 'random.uniform(2,5)'  # constant
config['get_output_time'] = '0.0'
config['job_id_resolved_time'] = 'random.uniform(10,50)'  # up to 800s
config['job_finish_time'] = '10+random.uniform(10,10)'
config['job_failure_rate'] = 'random.uniform(0,0.05)'

# submit K parallel master jobs with N subjobs each


def submit(N, K):
    jobs = []
    for i in range(K):
        j = Job()
        j._auto__init__()
        j.backend = LCG()
        j.backend.middleware = 'GLITE'
        j.splitter = GenericSplitter()
        j.splitter.attribute = 'application.args'
        j.splitter.values = [['x']] * N
        j.submit()
        jobs.append(j)
    import time

    def finished():
        for j in jobs:
            if not j.status in ['failed', 'completed']:
                return False
        return True

    while not finished():
        time.sleep(1)

    return jobs

# repeat M times for better statistics (and repository scalability)

M = 5

for i in range(M):
    logger.info('*' * 80)
    logger.info('starting %d out of %d' % (i, M))
    logger.info('*' * 80)
    submit(50, 10)

logger.info('finished!')



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import copy
import os
import pickle
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from RTHUtils import is_gaudi_child, getXMLSummaryScript, create_runscript
from Ganga.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from Ganga.GPIDev.Lib.File import FileBuffer, LocalFile, MassStorageFile
from Ganga.GPIDev.Base.Proxy import addProxy
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from Ganga.Utility.util import unique
from GangaGaudi.Lib.RTHandlers.GaudiRunTimeHandler import GaudiRunTimeHandler
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import script_generator, get_share_path, master_sandbox_prepare, sandbox_prepare
logger = getLogger()
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class LHCbGaudiRunTimeHandler(GaudiRunTimeHandler):

    """This is the application runtime handler class for Gaudi applications 
    using the local, interactive and LSF backends."""

    def master_prepare(self, app, appmasterconfig):
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig, ['inputsandbox'])

        # add summary.xml
        outputsandbox += ['summary.xml', '__parsedxmlsummary__']

        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):

        logger.debug("Prepare")

        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)

        job = app.getJobObject()

        logger.debug("Loading pickle files")

        #outputfiles=set([file.namePattern for file in job.outputfiles]).difference(set(getOutputSandboxPatterns(job)))
        # Cant wait to get rid of this when people no-longer specify
        # inputdata in options file
        #######################################################################
        # splitters ensure that subjobs pick up inputdata from job over that in
        # optsfiles but need to take sare of unsplit jobs
        if not job.master:
            share_path = os.path.join(get_share_path(app),
                                      'inputdata',
                                      'options_data.pkl')

            if not job.inputdata:
                if os.path.exists(share_path):
                    f = open(share_path, 'r+b')
                    job.inputdata = pickle.load(f)
                    f.close()

        #######################################################################
        # Cant wait to get rid of this when people no-longer specify
        # outputsandbox or outputdata in options file
        #######################################################################
        share_path = os.path.join(get_share_path(app),
                                  'output',
                                  'options_parser.pkl')

        logger.debug("Adding info from pickle files")

        if os.path.exists(share_path):
            f = open(share_path, 'r+b')
            parser = pickle.load(f)
            f.close()

            outbox, outdata = parser.get_output(job)

            from Ganga.GPIDev.Lib.File import FileUtils
            from Ganga.GPIDev.Base.Filters import allComponentFilters

            fileTransform = allComponentFilters['gangafiles']
            job.non_copyable_outputfiles.extend([fileTransform(this_file, None) for this_file in outdata if not FileUtils.doesFileExist(this_file, job.outputfiles)])
            job.non_copyable_outputfiles.extend([fileTransform(this_file, None) for this_file in outbox if not FileUtils.doesFileExist(this_file, job.outputfiles)])

            outputsandbox.extend([f.namePattern for f in job.non_copyable_outputfiles])

            outputsandbox.extend([f.namePattern for f in job.outputfiles])
            outputsandbox = unique(outputsandbox)
        #######################################################################

        logger.debug("Doing XML Catalog stuff")

        data = job.inputdata
        data_str = ''
        if data:
            logger.debug("Returning options String")
            data_str = data.optionsString()
            if data.hasLFNs():
                logger.debug("Returning Catalogue")
                inputsandbox.append(
                    FileBuffer('catalog.xml', data.getCatalog()))
                cat_opts = '\nfrom Gaudi.Configuration import FileCatalog\nFileCatalog().Catalogs = ["xmlcatalog_file:catalog.xml"]\n'
                data_str += cat_opts

        logger.debug("Doing splitter_data stuff")
        if hasattr(job, '_splitter_data'):
            data_str += job._splitter_data
        inputsandbox.append(FileBuffer('data.py', data_str))

        logger.debug("Doing GaudiPython stuff")

        cmd = 'python ./gaudipython-wrapper.py'
        opts = ''
        if is_gaudi_child(job.application):
            opts = 'options.pkl'
            cmd = 'gaudirun.py ' + \
                ' '.join(job.application.args) + ' %s data.py' % opts

        logger.debug("Setting up script")

        script = script_generator(create_runscript(job.application.newStyleApp),
                                  remove_unreplaced=False,
                                  OPTS=opts,
                                  PROJECT_OPTS=job.application.setupProjectOptions,
                                  APP_NAME=job.application.appname,
                                  APP_VERSION=job.application.version,
                                  APP_PACKAGE=job.application.package,
                                  PLATFORM=job.application.platform,
                                  CMDLINE=cmd,
                                  XMLSUMMARYPARSING=getXMLSummaryScript())  # ,
                                  # OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, ''))

        logger.debug("Returning StandardJobConfig")

        return StandardJobConfig(FileBuffer('gaudi-script.py', script, executable=1),
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



# Setup logging ---------------
from Ganga.Utility.logging import getLogger
log = getLogger()


class MonitoringClient(object):

    def __init__(self, monitoringService, serviceName='Main'):
        self.__MC = {}
        self.subscribe(serviceName, monitoringService)

    def subscribe(self, serviceName, monitoringService):
        if self.isSubscribed(serviceName):
            log.debug(
                "The %s service already exists. Please unsubscribe first." % serviceName)
            return False
        else:
            log.debug("Subscribing to the %s service." % serviceName)
        try:
            self.__MC[serviceName] = monitoringService
        except Exception as msg:
            log.debug(msg)
            return False
        return True

    def unsubscribe(self, serviceName='Main'):
        if self.isSubscribed(serviceName):
            del self.__MC[serviceName]
            return True
        return False

    def isSubscribed(self, serviceName='Main'):
        _s = serviceName in self.__MC
        if not _s:
            log.debug("Service %s does not exist." % serviceName)
        return _s

    def allStop(self):
        for service in self.__MC.keys():
            self.__MC[service].stop()

    def stop(self, serviceName='Main'):
        """Stops the disconnects from server and stops monitoring mechanism is required."""
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].stop()

    def pause(self, enableMonitoring=False, serviceName='Main'):
        if self.isSubscribed(serviceName):
            if enableMonitoring:
                self.__MC[serviceName].enableMonitoring()
            else:
                self.__MC[serviceName].disableMonitoring()

    def update(self, serviceName='Main'):
        self.__MC[serviceName].updateJobs()

    def combinedUpdate(self):
        for service in self.__MC:
            self.update(serviceName=service)

    def _getRegistry(self, serviceName='Main'):
        return self.__MC[serviceName].registry

    def _getUpdateJobStatusFunction(self, serviceName='Main'):
        return self.__MC[serviceName].updateJobStatus

    def makeUpdateJobStatusFunction(self, func, serviceName='Main'):
        return self.__MC[serviceName].makeUpdateJobStatusFunction(func)

    # Client callback functions

    def bindClientFunction(self, func, hookFunc, serviceName='Main'):
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].setClientCallback(func, hookFunc)

    def unbindClientFunction(self, func, serviceName='Main'):
        if self.isSubscribed(serviceName):
            try:
                self.__MC[serviceName].removeClientCallback(func)
            except Exception as msg:
                log.debug("unbindClientFunction() failed on %s: %s." %
                          (serviceName, msg))

    # Monitoring loop hook functions

    def bindMLFunction(self, hookFunc, argDict, enabled=True, serviceName='Main'):
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].setCallbackHook(hookFunc, argDict, enabled)

    def unbindMLFunction(self, hookFunc, serviceName='Main'):
        if self.isSubscribed(serviceName):
            try:
                self.__MC[serviceName].removeCallbackHook(hookFunc)
            except Exception as msg:
                log.debug("unbindClientFunction() failed on %s: %s." %
                          (serviceName, msg))

    # Monitor filters. NOT IN USE YET.

    def addFilter(self, mcFilterName, mcFilter, serviceName='Main'):
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].addFilter(mcFilterName, mcFilter)

    def removeFilter(self, mcFilterName, serviceName='Main'):
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].removeFilter(mcFilterName)

    def enableFilter(self, mcFilterName, enabled, serviceName='Main'):
        if self.isSubscribed(serviceName):
            self.__MC[serviceName].enableFilter(mcFilterName, enabled)

    # set attribute value on service
#    def setMCAttribute( self, attributeName, attributeValue, serviceName = 'Main' ):
#        if self.isSubscribed( serviceName ) and attributeName in [ 'gridProxy' ]:
#            setattr( self.__MC[ serviceName ], attributeName, attributeValue )

    def getMCAttribute(self, attributeName, serviceName='Main'):
        # and attributeName in [ 'gridProxy' ]:
        if self.isSubscribed(serviceName):
            return getattr(self.__MC[serviceName], attributeName)



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IMerger.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################
from Ganga.GPIDev.Adapters.IPostProcessor import PostProcessException, IPostProcessor
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base.Proxy import getName
import os
import glob

from Ganga.Utility.logging import getLogger
logger = getLogger()


class IChecker(IPostProcessor):

    """
    Abstract class which all checkers inherit from.
    """
    _schema = Schema(Version(1, 0), {
        'checkSubjobs': SimpleItem(defvalue=True, doc='Run on subjobs'),
        'checkMaster': SimpleItem(defvalue=True, doc='Run on master')
    })
    _category = 'postprocessor'
    _name = 'IChecker'
    _hidden = 1
    order = 2

    def execute(self, job, newstatus):
        """
        Execute the check method, if check fails pass the check and issue an ERROR message. Message is also added to the debug folder.
        """
        if newstatus == 'completed':
            #   If we're master job and check master check.
            #   If not master job and check subjobs check
            if (job.master is None and self.checkMaster) or\
                    ((job.master is not None) and self.checkSubjobs):
                try:
                    return self.check(job)
                except Exception as e:
                    with open(os.path.join(job.getDebugWorkspace().getPath(), 'checker_errors.txt'), 'a') as debug_file:
                        debug_file.write('\n Checker has failed with the following error: \n')
                        debug_file.write(str(e))
                    logger.error("%s" % e)
                    return True
        else:
            return True

    def check(self, job):
        """
        Method to check the output of jobs.
        Should be overidden.
        """
        raise NotImplementedError


class IFileChecker(IChecker):

    """
    Abstract class which all checkers inherit from.
    """
    _schema = IChecker._schema.inherit_copy()
    _schema.datadict['files'] = SimpleItem(defvalue=[], doc='File to search in')
    _schema.datadict['filesMustExist'] = SimpleItem(True, doc='Toggle whether to fail job if a file isn\'t found.')
    _category = 'postprocessor'
    _name = 'IFileChecker'
    _hidden = 1
    result = True
    order = 2

    def findFiles(self, job):

        if not len(self.files):
            raise PostProcessException(
                'No files specified, %s will do nothing!' % getName(self))

        filepaths = []
        for f in self.files:
            filepath = os.path.join(job.outputdir, f)
            for expanded_file in glob.glob(filepath):
                filepaths.append(expanded_file)
            if not len(glob.glob(filepath)):
                if (self.filesMustExist):
                    logger.info(
                        'The files %s does not exist, %s will fail job(%s) (to ignore missing files set filesMustExist to False)', filepath, self._name, job.fqid)
                    self.result = False
                else:
                    logger.warning(
                        'Ignoring file %s as it does not exist.', filepath)
        return filepaths




#Clean

from GangaRobot.Framework.Action import IAction
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
import os, shutil
from os.path import join
from GangaRobot.Framework.exceptions import *

logger = getLogger()

class Cleaner(IAction):
    """
    Cleaner IAction implementation
    """
    
    def execute(self, runid):
        #Clean up
        self._getconfig()
        logger.info("Starting Cleanup")
        self._clean()
        
    def _getconfig(self):
        config = getConfig('Configuration')
        self.releasedir = join(config['gangadir'],"Releases")
        #config = getConfig('IndependantTest')
        #self.ReleaseNumber = config['ReleaseNo']
        
    def _clean(self):
        #Removes the installed files
        installdir = join(self.releasedir,"install")
        logger.info("Removing install directory %s" % str(installdir))
        try:
            shutil.rmtree(installdir)
        except:
            raise GangaRobotContinueError("Failed to remove install directory",str(installdir))
        externaldir = join(self.releasedir,"external")
        logger.info("Removing external directory %s " % str(externaldir))
        try:
            shutil.rmtree(externaldir)
        except:
            raise GangaRobotContinueError("Failed to remove external directory",str(externaldir))
        tarballdir = join(self.releasedir,"tarball")
        logger.info("Removing tarball directory %s"  %str(tarballdir))
        try:
            shutil.rmtree(tarballdir)
        except:
            raise GangaRobotContinueError("Failed to remove tarball directory",str(tarballdir))
        """    
        jobdir = join(self.releasedir,"Jobs")
        logger.info("cleaning job directory %s"  %str(jobdir))
        try:
            shutil.rmtree(jobdir)
        except:
            raise GangaRobotContinueError("Failed to remove job directory",str(jobdir))
        """    
        



from __future__ import absolute_import
from .Config import getConfig, makeConfig, ConfigError, setSessionValuesFromFiles, allConfigs, setConfigOption, expandConfigPath, config_scope, setSessionValue
import os.path

## from Config import getConfigDict

# here are some useful option filters


def expandvars(c, v):
    """The ~ and $VARS are automatically expanded. """
    return os.path.expanduser(os.path.expandvars(v))


def expandgangasystemvars(c, v):
    """Expands vars with the syntax '@{VAR}' from the System config item."""
    system = getConfig('System')
    for key in system.options.iterkeys():
        option = '@{%s}' % key
        if option in v:
            v = v.replace(option, system[key])
    return v

def load_config_files():
    """
    Load the config files as a normal Ganga session would, taking
    into account environment variables etc.
    """
    from Ganga.Utility.Config import getConfig, setSessionValuesFromFiles
    from Ganga.Runtime import GangaProgram
    system_vars = {}
    for opt in getConfig('System'):
        system_vars[opt] = getConfig('System')[opt]
        user_config = os.environ.get('GANGA_CONFIG_FILE') or os.path.expanduser('~/.gangarc')
        config_files = GangaProgram.get_config_files(user_config)
        setSessionValuesFromFiles(config_files, system_vars)


def clear_config():
    """
    Reset all the configs back to their default values
    """
    from Ganga.Utility.Config import allConfigs
    for package in allConfigs.values():
        package._user_handlers = []
        package._session_handlers = []
        package.revertToDefaultOptions()




import inspect


def add_config(config_values):
    # type: (Sequence[Tuple[str, str, Any]])
    """
    A decorator to label functions with config arguments for
    consumption by the ``ganga`` fixture.

    Sets the ``_config_values`` attribute on the function.

    If passed a class it will label all the functions within. Config
    entries labelled on a function in a class will take precedence
    over those labelled on the class itself.

    Args:
        config_values: a list of 3-tuples (config section, entry name, entry value)

    Example:
        .. code-block:: python

            @add_config([('Queues', 'NumWorkerThreads', 7)])
            def test_add_config(ganga):
                assert ganga.config['Queues'].NumWorkerThreads == 7
    """
    # Convert the list of tuples to a dict for easy updating and comparison
    config_map = dict(((section, name), value) for section, name, value in config_values)

    def add_config_wrapper(obj):
        if inspect.isfunction(obj):
            obj._config_values = getattr(obj, '_config_values', {})  # Set the value to a dict if it doesn't exist

            # Since function decorators are called before class
            # decorators, existing config items take precedence
            for new_item, new_value in config_map.items():
                if new_item not in obj._config_values:
                    obj._config_values[new_item] = new_value

        else:  # We're on a class object
            for item in obj.__dict__.values():
                if inspect.isfunction(item):  # Only add config to functions
                    add_config_wrapper(item)
        return obj

    return add_config_wrapper


@add_config([('Configuration', 'user', 'test')])
def test_add_config(gpi):
    assert gpi.config['Configuration'].user == 'test'


@add_config([('Configuration', 'user', 'test'), ('MSGMS', 'port', 43)])
class TestLayeredConfig(object):

    def test_class_config(self, gpi):
        assert gpi.config['Configuration'].user == 'test'
        assert gpi.config['MSGMS'].port == 43

    @add_config([('MSGMS', 'port', 42)])
    def test_config_override(self, gpi):
        assert gpi.config['Configuration'].user == 'test', 'Function-decorated configs should not remove all class ones'
        assert gpi.config['MSGMS'].port == 42, 'Function-decorated configs should take precedence over class ones'



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from Ganga.GPIDev.Adapters.IPostProcessor import PostProcessException
from Ganga.GPIDev.Adapters.IChecker import IChecker
from Ganga.GPIDev.Schema import SimpleItem
from Ganga.Utility.logging import getLogger
import commands
import copy
import os
import re


logger = getLogger()


class MetaDataChecker(IChecker):

    """
    Checks the meta data of a job This class must be overidden to convert the experiment specific metadata.
    """
    _schema = IChecker._schema.inherit_copy()
    _schema.datadict['expression'] = SimpleItem(
        defvalue=None, typelist=[str, None], doc='The metadata attribute')
    _schema.datadict['result'] = SimpleItem(
        defvalue=None, typelist=[bool, None], hidden=1, doc='Check result')
    _category = 'postprocessor'
    _name = 'MetaDataChecker'
    _hidden = 1

    def calculateResult(self, job):
        """
        To be overidden by experiment specific class
        """
        raise NotImplementedError

    def check(self, job):
        """
        Checks metadata of job is within a certain range.
        """
        if self.expression == None:
            raise PostProcessException('No expression is set. MetaDataChecker will do nothing!')
        try:
            self.result = self.calculateResult(job)
        except Exception as e:
            raise PostProcessException('There was an error parsing the checker expression: %s - MetaDataChecker will do nothing!' % e)
        if self.result is not True and self.result is not False:
            raise PostProcessException('The expression "%s" did not evaluate to True or False, MetaDataChecker will do nothing!' % self.expression)
        if self.result is False:
            logger.info('MetaDataChecker has failed job(%s) because the expression "%s" is False' % (job.fqid, self.expression))
        return self.result





import datetime
import time
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *
from Ganga.Core import BackendError
from Ganga.Lib.Batch import Batch

import Ganga.Utility.Config

from Ganga.Core import FileWorkspace
import os




config = Ganga.Utility.Config.makeConfig('WestGrid','internal WestGrid command line interface')

config.addOption('shared_python_executable', False, "Shared PYTHON")

config.addOption('jobid_name', 'PBS_JOBID', "Name of environment with ID of the job")
config.addOption('queue_name', 'PBS_QUEUE', "Name of environment with queue name of the job")
config.addOption('heartbeat_frequency', '30', "Heartbeat frequency config variable")

config.addOption('submit_str', 'cd %s; qsub %s %s %s %s', "String used to submit job to queue")
config.addOption('submit_res_pattern', '^(?P<id>\S*)', "String pattern for replay from the submit command")

config.addOption('stdoutConfig', '-o %s/stdout', "String pattern for defining the stdout")
config.addOption('stderrConfig', '-e %s/stderr', "String pattern for defining the stderr")

config.addOption('kill_str', 'qdel %s', "String used to kill job")
config.addOption('kill_res_pattern', '(^$)|(qdel: Unknown Job Id)', "String pattern for replay from the kill command")

tempstr='''
env = os.environ
jobnumid = env["PBS_JOBID"]
os.system("mkdir /tmp/%s/" %jobnumid)
os.chdir("/tmp/%s/" %jobnumid)
os.environ["PATH"]+=":."
'''
config.addOption('preexecute', tempstr, "String contains commands executing before submiting job to queue")

tempstr='''
env = os.environ
jobnumid = env["PBS_JOBID"]
os.chdir("/tmp/")
os.system("rm -rf /tmp/%s/" %jobnumid) 
'''
config.addOption('postexecute', tempstr, "String contains commands executing before submiting job to queue")
config.addOption('jobnameopt', 'N', "String contains option name for name of job in batch system")
config.addOption('timeout',600,'Timeout in seconds after which a job is declared killed if it has not touched its heartbeat file. Heartbeat is touched every 30s so do not set this below 120 or so.')

config.addOption('voproxy',None,'Path to your vo-proxy')

class WestGrid(Batch):
    ''' WestGrid backend - submit jobs to Portable Batch System.
    '''
    #same schema as a Batch, but with the added option of where the user proxy is stored
    _schema = Schema(Version(1,0), {'queue' : SimpleItem(defvalue='',doc='queue name as defomed in your local Batch installation'),
                                    'extraopts' : SimpleItem(defvalue='',doc='extra options for Batch. See help(Batch) for more details'),
                                    'id' : SimpleItem(defvalue='',protected=1,copyable=0,doc='Batch id of the job'),
                                    'exitcode' : SimpleItem(defvalue=None,typelist=['int','type(None)'],protected=1,copyable=0,doc='Process exit code'),
                                    'status' : SimpleItem(defvalue='',protected=1,hidden=1,copyable=0,doc='Batch status of the job'),
                                    'actualqueue' : SimpleItem(defvalue='',protected=1,copyable=0,doc='queue name where the job was submitted.'),
                                    'actualCE' : SimpleItem(defvalue='',protected=1,copyable=0,doc='hostname where the job is/was running.'),
                                    'voproxy' : SimpleItem(defvalue=None,typelist=['str','type(None)'],doc='select specific proxy location')
                                    })
    _category = 'backends'
    _name = 'WestGrid'

    config = Ganga.Utility.Config.getConfig('WestGrid')
    def __init__(self):
        super(WestGrid,self).__init__()
        if self.voproxy:
            print "USING MY OWN VO PROXY:",self.voproxy
        elif self.config['voproxy']:
            self.voproxy = os.path.expanduser(self.config['voproxy'])



from Ganga.GPIDev.Base import GangaObject
from Ganga.Core.exceptions import GangaAttributeError
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.GPIDev.Base.Proxy import addProxy


class MetadataDict(GangaObject):

    '''MetadataDict class

    Class that represents the dictionary of metadata.
    '''
    _schema = Schema(Version(1, 0), {
        'data': SimpleItem(defvalue={}, doc='dict data', hidden=1, protected=1)
    })
    _name = 'MetadataDict'
    _category = 'metadata'
    _exportmethods = ['__getitem__']  # ,'__str__']

    def __init__(self):
        super(MetadataDict, self).__init__()

    def __getitem__(self, key):
        import copy
        return addProxy(copy.deepcopy(self.data[key]))

    def __setitem__(self, key, value):
        from Ganga.GPIDev.Lib.Job.Job import Job
        if key in Job._schema.datadict:
            raise GangaAttributeError(
                '\'%s\' is a reserved key name and cannot be used in the metadata' % key)
        if not isinstance(key, str):
            raise GangaAttributeError(
                'Metadata key must be of type \'str\' not %s' % type(key))
        if isinstance(value, GangaObject):
            raise GangaAttributeError(
                'Metadata doesn\'t support nesting of GangaObjects at the moment')

        self.data[key] = value
        self._setDirty()

    def update(self, dict):

        # this way pick up the checking for free
        for key, value in dict.iteritems():
            self.__setitem__(key, value)
#        self.data.update(dict)

    def printSummaryTree(self, level=0, verbosity_level=0, whitespace_marker='', out=None, selection='', interactive=False):
        """If this method is overridden, the following should be noted:

        level: the hierachy level we are currently at in the object tree.
        verbosity_level: How verbose the print should be. Currently this is always 0.
        whitespace_marker: If printing on multiple lines, this allows the default indentation to be replicated.
                           The first line should never use this, as the substitution is 'name = %s' % printSummaryTree()
        out: An output stream to print to. The last line of output should be printed without a newline.'
        selection: See VPrinter for an explaintion of this.
        """
        if len(self.data) == 0:
            out.write('{}\n')
            return
        out.write('{\n')
        for key, value in self.data.iteritems():
            out.write(whitespace_marker + '     ' + str(key) + ' = ' + str(value) + '\n')
        out.write(whitespace_marker + '    }\n')



#! /usr/bin/env python
################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ganga-joboption-parse.py,v 1.1 2008-07-17 16:41:18 moscicki Exp $
################################################################
# Parse Athena jobOptions files

import os, sys, commands, time, re, math, random

# error codes
# WRAPLCG_UNSPEC
EC_UNSPEC        = 410000 
# WRAPLCG_STAGEOUT_UNSPEC
EC_STAGEOUT      = 410400

# Get output jobid
try:
    output_jobid = os.environ['OUTPUT_JOBID']
except:
    raise LookupError("ERROR: OUTPUT_JOBID not defined")
    sys.exit(EC_STAGEOUT)

try:
    output_datasetname = os.environ['OUTPUT_DATASETNAME']
except:
    raise LookupError("ERROR: OUTPUT_DATASETNAME not defined")
    sys.exit(EC_STAGEOUT)

# Get output jobid
try:
    use_short_filename = os.environ['GANGA_SHORTFILENAME']
except:
    raise LookupError("ERROR: GANGA_SHORTFILENAME not defined")
    sys.exit(EC_STAGEOUT)

# Get output jobid
try:
    dq2_outputfile_namelength = os.environ['DQ2_OUTPUTFILE_NAMELENGTH']
except:
    print LookupError("ERROR: DQ2_OUTPUTFILE_NAMELENGTH not defined - using 150 characters")
    dq2_outputfile_namelength = 150

    
# Read output_files 
try:
    open('output_files','r')
except IOError:
    print 'No output_files requested.'
    sys.exit(EC_UNSPEC)
output_files = [ line.strip() for line in file('output_files') ]
if not len(output_files):
    print 'No output_files requested.'
    sys.exit(EC_UNSPEC)
    
#Create output_files.new with new outputfilenames
new_output_files = {}
outFile = open('output_files.new','w')
for output_file in output_files:
    temptime = time.gmtime()
    output_datasetname = re.sub('\.[\d]+$','',output_datasetname)
    pattern=output_datasetname+".%04d%02d%02d%02d%02d%02d%04d._%05d."+output_file
    i=output_jobid.split('.')
    rannum = int(math.ceil(10000*random.uniform(0,1)))
    if len(i)>1:
        new_output_file = pattern % (temptime[0],temptime[1],temptime[2],temptime[3],temptime[4],temptime[5], rannum,int(i[1])+1)
        short_pattern = ".%04d%02d%02d%02d%02d%02d%04d._%05d" % (temptime[0],temptime[1],temptime[2],temptime[3],temptime[4],temptime[5], rannum,int(i[1])+1)
    else:
        new_output_file = pattern % (temptime[0],temptime[1],temptime[2],temptime[3],temptime[4],temptime[5], rannum,1)
        short_pattern = ".%04d%02d%02d%02d%02d%02d%04d._%05d" % (temptime[0],temptime[1],temptime[2],temptime[3],temptime[4],temptime[5], rannum,1)


    if len(new_output_file)>dq2_outputfile_namelength:
        print '!!!!!!!!! ERROR - JOB ABORTING !!!!!!!!!!!!!!!!!!!!!!!!!! '
        print '!!!!!!!!! Output filename length is larger than %s characters!!!!!!!!!!!!!!!!!' %dq2_outputfile_namelength
        print '!!!!!!!!! %s is too long !!!!!!!!!!!!!!!!!!!!' % new_output_file
        print '!!!!!!!!! Please use a shorter output dataset name !!!!!!!!!!!!!!!!!!'
        sys.exit(EC_STAGEOUT)
    
    new_short_output_file = re.sub(".root", short_pattern+".root" , output_file )
    if new_short_output_file == output_file:
        new_short_output_file =  short_pattern[1:] + "." + output_file

    if use_short_filename:
        new_output_files[output_file] = new_short_output_file
        outFile.write(new_short_output_file)
    else:
        new_output_files[output_file] = new_output_file
        outFile.write(new_output_file)

    outFile.write('\n')
outFile.close()
       
# Parse jobOptions file to append jobid to output_files
if 'ATHENA_OPTIONS' in os.environ:
    joboptions = os.environ['ATHENA_OPTIONS'].split(' ')

    for jfile in joboptions:
        try:
            open(jfile,'r')
        except IOError:
            continue
        for output_file in output_files:
            cmd = "sed 's/"+output_file+"/"+new_output_files[output_file]+"/' "+jfile+" > " + jfile+".new"
            rc, output = commands.getstatusoutput(cmd)
            if rc:
                print "ERROR: parsing output file in jobOptions"
                print cmd
                print rc, output
            os.chmod(jfile+".new", 0o777)
            os.rename(jfile+".new", jfile)   



from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Schema import *
from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Lib.File import File
from Ganga.Core.exceptions import ApplicationConfigurationError
import os

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

class PandaPilot(IApplication):
    """
    PandaPilot application -- start a Panda Pilot on a remote CE
    
    """
    _schema = Schema(Version(1,0), {
        'exe' : SimpleItem(defvalue=File(os.path.join(os.path.dirname(__file__),'runpilot3-script-stub.sh')),typelist=['Ganga.GPIDev.Lib.File.File.File'],doc='File object containing the pilot startup script. Do not modify unless you know what you are doing.'), 
        'arg_pattern' : SimpleItem(defvalue='-s %%QUEUE%% -h %%QUEUE%% -j false -u self',typelist=['str'],doc="Pilot argument pattern. Do not modify unless you know what you are doing."),
        'queue' : SimpleItem(defvalue='',typelist=['str'],doc='The analysis queue to register the pilot on e.g. ANALY_TRIUMF')
        })
    _category = 'applications'
    _name = 'PandaPilot'
    _GUIPrefs = []
    _GUIAdvancedPrefs = []

    def __init__(self):
        super(PandaPilot,self).__init__()
        
    def configure(self,masterappconfig):
        return (None,None)

config = getConfig('defaults_PandaPilot') #_Properties
config.options['exe'].type = type(None)

def queueToCE(queue):
    from pandatools import Client
    return Client.PandaSites[queue]['queue']+'-'+Client.PandaSites[queue]['localqueue']

def queueToDDM(queue):
    from pandatools import Client
    return Client.PandaSites[queue]['ddm']

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from Ganga.Lib.LCG import LCGJobConfig
        if not app.queue:
            raise ApplicationConfigurationError(None,'queue not specified for PandaPilot')
        args = app.arg_pattern.replace('%%QUEUE%%',app.queue)
        job = app._getParent()

        if not job.backend.requirements.sites:
            job.backend.requirements.sites=[queueToDDM(app.queue)]
            if not job.backend.requirements.sites:
                raise ApplicationConfigurationError(None,'Could not map queue name to LCG site')
        logger.info('Sending pilot for %s to %s'%(app.queue,job.backend.requirements.sites))
        return LCGJobConfig(app.exe,app._getParent().inputsandbox,[args],app._getParent().outputsandbox,None)

from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('PandaPilot','LCG', LCGRTHandler)



from Ganga.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaLHCb.Lib.RTHandlers.LHCbGaudiRunTimeHandler import LHCbGaudiRunTimeHandler


def addLocalTestSubmitter():
    allHandlers.add('DaVinci', 'TestSubmitter', LHCbGaudiRunTimeHandler)



#
# Dictionary with all parameters available for the
# section CMSSW.
#
# 08/06/10 @ ubeda
#

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

class CMSSW(GangaObject):

    _comments = []
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#datasetpath__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#runselection__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#use_parent')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#pset__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#pycfg_params__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#lumi_mask')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#total_number_of_events__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#events_per_job_')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#total_number_of_lumis__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#lumis_per_job_')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#number_of_jobs__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#split_by_run')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#output_file__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#skip_tfileservice_output')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#get_edm_output')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#increment_seeds')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#preserve_seeds')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#first_lumi')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#generator')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#executable')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#dbs_url')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#show_prod')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#subscribed')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#no_block_boundary')
    _comments.append('--- TO DO ---')

    schemadic = {}
    schemadic['datasetpath']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[0])
    schemadic['runselection']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[1])
    schemadic['use_parent']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[2])
    schemadic['pset']                     = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[3])
    schemadic['pycfg_params']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[4])
    schemadic['lumi_mask']                = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[5])
    schemadic['total_number_of_events']   = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[6])
    schemadic['events_per_job']           = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[7])
    schemadic['split_by_event']           = SimpleItem(defvalue=None, typelist=['type(None)','int'])
    schemadic['total_number_of_lumis']    = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[8])
    schemadic['lumis_per_job']            = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[9])
    schemadic['number_of_jobs']           = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[10])
    schemadic['split_by_run']             = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[11])
    schemadic['output_file']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[12])
    schemadic['skip_TFileService_output'] = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[13])
    schemadic['get_edm_output']           = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[14])
    schemadic['increment_seeds']          = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[15])
    schemadic['preserve_seeds']           = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[16])
    schemadic['first_lumi']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[17])
    schemadic['generator']                = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[18])
    schemadic['executable']               = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[19])
    schemadic['dbs_url']                  = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[20])
    schemadic['show_prod']                = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[21])
    schemadic['subscribed']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[22])
    schemadic['no_block_boundary']        = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[23])
    schemadic['ignore_edm_output']        = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[24])
    schemadic['use_dbs_1']                = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[24])


    _schema =  Schema(Version(0,0), {})
    _hidden = 1



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PACKAGE.py,v 1.7 2009-07-27 15:15:56 moscicki Exp $
##########################################################################

""" PACKAGE modules describe the installation and setup of the Ganga runtime packages.
Purpose: automatic initialization of Ganga environment setup and software distribution tools.

Each PACKAGE module should provide:
 - setup object : an instance of Ganga.Utility.Setup.PackageSetup class
 - standardSetup() function

The setup object is used to describe the external dependencies of the package.
The standardSetup() function is used to perform automatic initialization of the environment of the package.
"""

# Default minimum Python version number asked for by Ganga
_defaultMinVersion = "2.6"
_defaultMinHexVersion = 0x20600f0

# The default values will be guessed but you may override them here
_defaultPlatform = None  # 'slc3_gcc323'
_defaultExternalHome = None
#_defaultExternalHome = "/afs/cern.ch/sw/ganga/external/"

# The dictionary of external  packages is used by the release/download
# system to handle the installation  tarballs. Make sure that all your
# dependencies go in here
#
# The layout of the external tree is fixed. The package top is:
#  externalHome/name/version/platf
#
# This layout is similar to  LCG external software repository and thus
# may be profitable in the long term.
#
# syspath specifies the relative path for PYTHONPATH setup (sys.path)
# if noarch=1 then platf is ommited
#
# if a path-like variable (e.g. LD_LIBRARY_PATH) is set using setup.prependPath() method, the
# value specified in the dictionary may be either string or a list of strings (they will be separated by colons ':').
#
_externalPackages = {
    'ipython': {'version': '1.2.1',
                'noarch': True,
                'syspath': 'lib/python'},
#    'paramiko': {'version': '1.7.3',
#                 'noarch': True,
#                 'syspath': 'lib/python2.3/site-packages'},
#    'pycrypto': {'version': '2.0.1',
#                 'syspath': 'lib/python2.3/site-packages'},
    'stomputil': {'version': '2.4',
                  'noarch': True,
                  'syspath': 'python'},
    'httplib2': {'version': '0.8',
                 'noarch': True,
                 'syspath': 'python'},
    'python-gflags': {'version': '2.0',
                      'noarch': True,
                      'syspath': 'python'},
    'google-api-python-client': {'version': '1.1',
                                 'noarch': True,
                                 'syspath': 'python'}
}


def detectPlatform():
    """ Try to guess the platform string according to the operating system, current environment and python interpreter.
    Ganga provides precompiled external packages on a limited set of _default platforms_ as explained in:
    https://twiki.cern.ch/twiki/bin/view/ArdaGrid/GangaSupportedPlatforms
    This function is set only to detect the well-known platform strings as defined by the LCG SPI project and is not meant to be a
    generic platform detection utility. If the platform cannot be guessed a default one is returned. This may or may not work on 
    other systems. In this case you should resolve the external binary dependencies yourself.

    Comments about current implementations:

    SLC5 platform is detected using platform module.

    If it's not SLC5 then:

    We assume that 64 bit python implies the slc4, amd64 system.
    We assume that 32 bit python implies the slc4, ia32 system.

    We ignore IA64 architecture (Opteron) as not frequently used.

    """

    # assume INTEL processors (i386, i686,x64), ignore IA64 architecture
    platf4 = {32: 'slc4_ia32_gcc34', 64: 'slc4_amd64_gcc34'}
    platf5 = {32: 'i686-slc5-gcc43-opt', 64: 'x86_64-slc5-gcc43-opt'}

    # for older python versions use some tricks
    import sys
    bits = sys.maxsize >> 32

    if bits:
        arch = 64
    else:
        arch = 32

    platfstring = platf4

    import platform
    import re
    c = re.compile('\S+-redhat-(?P<ver>\S+)-\S+')
    r = c.match(platform.platform())
    if r and r.group('ver').split('.')[0] == '5':
        platfstring = platf5

    return platfstring[arch]


# guess defaults if not defined
if not _defaultExternalHome:
    import os.path
    import Ganga
    from Ganga.Utility.files import fullpath
    p = fullpath(Ganga.__file__)
    for i in range(5):
        p = os.path.dirname(p)
    _defaultExternalHome = os.path.join(p, 'external')

if not _defaultPlatform:
    _defaultPlatform = detectPlatform()


from Ganga.Utility.Setup import PackageSetup
# The setup object
setup = PackageSetup(_externalPackages)


def standardSetup(setup=setup):
    """ Perform automatic initialization of the environment of the package.
    The gangaDir argument is only used by the core package, other packages should have no arguments.
    """

    from Ganga.Utility.Setup import checkPythonVersion
    import sys

    # here we assume that the Ganga has been already prepended to sys.path by
    # the caller
    if checkPythonVersion(_defaultMinVersion, _defaultMinHexVersion):
        for name in setup.packages:
#            if name == 'pycrypto' and sys.hexversion > 0x2050000:
#                # hack the pycrypto path for 2.5
#                setup.packages['pycrypto']['syspath'] = setup.packages['pycrypto']['syspath'].replace('2.3', '2.5')
#
#            if name == 'paramiko' and sys.hexversion > 0x2050000:
#                # hack the paramiko path for 2.5
#                setup.packages['paramiko']['syspath'] = setup.packages['paramiko']['syspath'].replace('2.3', '2.5')

            setup.setSysPath(name)
            setup.prependPath(name, 'PYTHONPATH')

            # if other PATH variable must be defined, e.g. LD_LIBRARY_PATH, then
            # you should do it this way:
            # setup.prependPath(name,'LD_LIBRARY_PATH')
    else:
        sys.exit()




"""Dirac-specific extractor IAction implementation.

The DiracExtractor class extracts data specific to the Dirac backend.

N.B. Since only a few data are extracted, this class should currently be
considered only as an example of a backend-specific extractor. More complete and
relevant data may be extracted in a later version.
 
"""

from GangaRobot.Lib.Base.BaseExtractor import BaseExtractor
from Ganga.GPI import *
from GangaLHCb.Lib.DIRAC.DiracUtils import *
#from GangaLHCb.Lib.DIRAC.DiracServer import DiracServer
from GangaDirac.Lib.Utilities.DiracUtilities import execute

configLHCb = Ganga.Utility.Config.getConfig('LHCb')
configDirac = Ganga.Utility.Config.getConfig('DIRAC')
#dirac_ganga_server = DiracServer()

logger = Ganga.Utility.logging.getLogger()


class DiracExtractor(BaseExtractor):
    
    """Dirac-specific extractor IAction implementation.
    
    An extractor which extracts some data specific to the Dirac backend for
    Ganga jobs using Dirac as the backend. See handlejobnode() for details.
    
    The extracted run node has the structure:
    <run>
        <job>
            <dirac/>
        </job>
        <job>
            <dirac/>
        </job>
        ...
    </run>

    This extractor can be reused by adding it to the chain of extractors
    initialised in the constructor of any implementation of BaseExtractor.
    e.g.
    def __init__(self):
        self.chain = [CoreExtractor(), DiracExtractor(), self]
    
    """

    def handlejobnode(self, jobnode, job):
        """Extracts data specific to the Dirac backend for the given job.
        
        Keyword arguments:
        jobnode -- An Extract.Node for the given job.
        job -- A Ganga Job associated with the run.
        
        If the job uses the Dirac backend then the following nodes are added to
        the job node:
        <dirac>
            <status/>
            <application-status/>
            <edg-wl-jobid/>
            <cpu/>
        </dirac>
        where the subnodes of the dirac node contain text values from querying
        the getJobSummary and getJobParams methods of DIRAC.Client.Dirac.Dirac.
        
        Subnodes are guaranteed to be present although their text values may be
        empty if querying Dirac returns no corresponding value.
        
        """
        if job.backend._impl._name != 'Dirac':
            return
        diracnode = jobnode.addnode('dirac')
        # data to extract
        status = None
        applicationstatus = None
        edgwljobid = None
        cpu = None
        # extract data
        diracid = job.backend.id
        if job.backend._impl._name != 'Dirac':
                        return
        try:
                command='output( dirac.getJobSummary([%i]) )'%diracid
                result=execute(command)
                #FIXME missing error checking

                jobSummary=result['Value'][diracid]
                applicationstatus = jobSummary['ApplicationStatus']
                
                command="output( dirac.getJobCPUTime(%i) )"%diracid
                cputime=result['Value'][diracid]
                cpu=cputime.get('CPUConsumed',None)
                #FIXME fins example time
                

               
        except KeyError:
                pass # data unavailable
        # add subnodes
        diracnode.addnode('status', status)
        diracnode.addnode('application-status', applicationstatus)
        diracnode.addnode('cpu', cpu)




"""
Stack tracer for multi-threaded applications.

Usage:

from Ganga.Utility import stacktracer
stacktracer.trace_start('trace.html')
....
stacktracer.trace_stop()
"""

import os
import sys
import threading
import time
import traceback
from datetime import datetime

try:
    from pygments import highlight
    from pygments.formatters.html import HtmlFormatter
    from pygments.lexers.python import Python3TracebackLexer
    def highlight_trace(string):
        return highlight(string, Python3TracebackLexer(), HtmlFormatter(noclasses=True))
except ImportError:
    def highlight_trace(string):
        return '<pre>' + string + '</pre>'

from Ganga.Utility.Config import getConfig


def stacktraces():
    # type: () -> str
    """
    Based on http://bzimmer.ziclix.com/2008/12/17/python-thread-dumps/

    Returns:
        str: HTML code to be saved to a file
    """
    html = [
        '<!doctype html>',
        '<html lang="en">',
        '<head>',
        '<meta charset=utf-8>',
        '<title>Current Ganga Threads</title>',
        '</head>'
        '<body>'
        '<h1>Current Ganga Threads</h1>',
    ]
    for thread_id, stack in sys._current_frames().items():
        name = dict((t.ident, t.name) for t in threading.enumerate())
        title = '<h2>{0}</h2>'.format(name.get(thread_id, None))
        html.append(title)

        trace = traceback.format_stack(stack)
        html.append(highlight_trace(''.join(trace)))

    html.append('<small>' + datetime.utcnow().isoformat() + '</small>')
    html.append('</body>')
    html.append('</html>')

    return '\n'.join(html)


class TraceDumper(threading.Thread):
    """
    Dump stack traces into a given file periodically.

    This part was heavily based on code by
    `nagylzs <http://code.activestate.com/recipes/577334-how-to-debug-deadlocked-multi-threaded-programs/>`_
    """
    def __init__(self, path, interval, auto):
        # type: (str, int, bool) -> None
        """
        Args:
            path: File path to output HTML (stack trace file)
            interval: In seconds: how often to update the trace file.
            auto: Set flag (True) to update trace continuously.
                Clear flag (False) to update only if file not exists.
                (Then delete the file to force update.)
        """
        assert interval > 0.1, 'Stacktracer interval must be greater than 0.1 seconds'
        threading.Thread.__init__(self)
        self.auto = auto
        self.interval = interval
        self.path = os.path.abspath(path)
        self.stop_requested = threading.Event()
        self.daemon = True

    def run(self):
        while not self.stop_requested.isSet():
            time.sleep(self.interval)
            if self.auto or not os.path.isfile(self.path):
                self.stacktraces()

    def stop(self):
        self.stop_requested.set()
        self.join(timeout=0)
        try:
            os.unlink(self.path)
        except OSError:
            pass

    def stacktraces(self):
        try:
            with open(self.path, 'wb+') as fout:
                fout.write(stacktraces())
        except IOError:
            pass  # Don't warn if the file faile to write


_tracer = None


def trace_start(filename='thread_trace.html', interval=5, auto=True):
    # type: (str, int, bool) -> None
    """
    Start tracing into the given file.

    Args:
        path: File path to output HTML (stack trace file)
        interval: In seconds: how often to update the trace file.
        auto: Set flag (True) to update trace continuously.
            Clear flag (False) to update only if file not exists.
            (Then delete the file to force update.)
    """
    path = os.path.join(getConfig('Configuration')['gangadir'], filename)
    global _tracer
    if _tracer is None:
        _tracer = TraceDumper(path, interval, auto)
        _tracer.start()
    else:
        raise Exception('Already tracing to {0}'.format(_tracer.path))


def trace_stop():
    """Stop tracing."""
    global _tracer
    if _tracer is None:
        raise Exception('Not tracing, cannot stop.')
    else:
        _tracer.stop()
        _tracer = None



#!/usr/bin/env python

import math

def floatRgb(mag, cmin, cmax):
    """
    These functions, when given a magnitude mag between cmin and cmax, return
    a colour tuple (red, green, blue). Light blue is cold (low magnitude)
    and yellow is hot (high magnitude).

    Return a tuple of floats between 0 and 1 for the red, green and
    blue amplitudes.
    """
    try:
        # normalize to [0,1]
        x = float(mag-cmin)/float(cmax-cmin)
    except:
        # cmax = cmin
        x = 0.5
    blue = min((max((4*(0.75-x), 0.)), 1.))
    red  = min((max((4*(x-0.25), 0.)), 1.))
    green= min((max((4*math.fabs(x-0.5)-1., 0.)), 1.))
    return (red, green, blue)

def rgb(mag, cmin, cmax):
    """
    Return a tuple of integers to be used in AWT/Java plots.
    """
    red, green, blue = floatRgb(mag, cmin, cmax)
    return (int(red*255), int(green*255), int(blue*255))

# function for mapping tld to country name
def tld_country(tld):
    tld_countries = {
       "af": "Afghanistan",
       "al": "Albania",
       "dz": "Algeria",
       "as": "American Samoa",
       "ad": "Andorra",
       "ao": "Angola",
       "ai": "Anguilla",
       "aq": "Antarctica",
       "ag": "Antigua and Barbuda",
       "ar": "Argentina",
       "am": "Armenia",
       "aw": "Aruba",
       "au": "Australia",
       "at": "Austria",
       "az": "Azerbaijan",
       "bs": "Bahamas",
       "bh": "Bahrain",
       "bd": "Bangladesh",
       "bb": "Barbados",
       "by": "Belarus",
       "be": "Belgium",
       "bz": "Belize",
       "bj": "Benin",
       "bm": "Bermuda",
       "bt": "Bhutan",
       "bo": "Bolivia",
       "ba": "Bosnia and Herzegowina",
       "bw": "Botswana",
       "bv": "Bouvet Island",
       "br": "Brazil",
       "io": "British Indian Ocean Territory",
       "bn": "Brunei Darussalam",
       "bg": "Bulgaria",
       "bf": "Burkina Faso",
       "bi": "Burundi",
       "kh": "Cambodia",
       "cm": "Cameroon",
       "ca": "Canada",
       "cv": "Cape Verde",
       "ky": "Cayman Islands",
       "cf": "Central African Republic",
       "td": "Chad",
       "cl": "Chile",
       "cn": "China",
       "cx": "Christmas Island",
       "cc": "Cocos (Keeling) Islands",
       "co": "Colombia",
       "km": "Comoros",
       "cg": "Congo",
       "cd": "Congo, The Democratic Republic of the",
       "ck": "Cook Islands",
       "cr": "Costa Rica",
       "ci": "Cote D'Ivoire",
       "hr": "Croatia",
       "cu": "Cuba",
       "cy": "Cyprus",
       "cz": "Czech Republic",
       "dk": "Denmark",
       "dj": "Djibouti",
       "dm": "Dominica",
       "do": "Dominican Republic",
       "tp": "East Timor",
       "ec": "Ecuador",
       "eg": "Egypt",
       "sv": "El Salvador",
       "gq": "Equatorial Guinea",
       "er": "Eritrea",
       "ee": "Estonia",
       "et": "Ethiopia",
       "fk": "Falkland Islands (Malvinas)",
       "fo": "Faroe Islands",
       "fj": "Fiji",
       "fi": "Finland",
       "fr": "France",
       "gf": "French Guiana",
       "pf": "French Polynesia",
       "tf": "French Southern Territories",
       "ga": "Gabon",
       "gm": "Gambia",
       "ge": "Georgia",
       "de": "Germany",
       "gh": "Ghana",
       "gi": "Gibraltar",
       "gr": "Greece",
       "gl": "Greenland",
       "gd": "Grenada",
       "gp": "Guadeloupe",
       "gu": "Guam",
       "gt": "Guatemala",
       "gn": "Guinea",
       "gw": "Guinea-Bissau",
       "gy": "Guyana",
       "ht": "Haiti",
       "hm": "Heard Island and Mcdonald Islands",
       "va": "Holy See (Vatican City State)",
       "hn": "Honduras",
       "hk": "Hong Kong",
       "hu": "Hungary",
       "is": "Iceland",
       "in": "India",
       "id": "Indonesia",
       "ir": "Iran, Islamic Republic of",
       "iq": "Iraq",
       "ie": "Ireland",
       "il": "Israel",
       "it": "Italy",
       "jm": "Jamaica",
       "jp": "Japan",
       "jo": "Jordan",
       "kz": "Kazakstan",
       "ke": "Kenya",
       "ki": "Kiribati",
       "kp": "Korea, Democratic People's Republic of",
       "kr": "Korea, Republic of",
       "kw": "Kuwait",
       "kg": "Kyrgyzstan",
       "la": "Lao People's Democratic Republic",
       "lv": "Latvia",
       "lb": "Lebanon",
       "ls": "Lesotho",
       "lr": "Liberia",
       "ly": "Libyan Arab Jamahiriya",
       "li": "Liechtenstein",
       "lt": "Lithuania",
       "lu": "Luxembourg",
       "mo": "Macau",
       "mk": "Macedonia",
       "mg": "Madagascar",
       "mw": "Malawi",
       "my": "Malaysia",
       "mv": "Maldives",
       "ml": "Mali",
       "mt": "Malta",
       "mh": "Marshall Islands",
       "mq": "Martinique",
       "mr": "Mauritania",
       "mu": "Mauritius",
       "yt": "Mayotte",
       "mx": "Mexico",
       "fm": "Micronesia",
       "md": "Moldova",
       "mc": "Monaco",
       "mn": "Mongolia",
       "ms": "Montserrat",
       "ma": "Morocco",
       "mz": "Mozambique",
       "mm": "Myanmar",
       "na": "Namibia",
       "nr": "Nauru",
       "np": "Nepal",
       "nl": "Netherlands",
       "an": "Netherlands Antilles",
       "nc": "New Caledonia",
       "nz": "New Zealand",
       "ni": "Nicaragua",
       "ne": "Niger",
       "ng": "Nigeria",
       "nu": "Niue",
       "nf": "Norfolk Island",
       "mp": "Northern Mariana Islands",
       "no": "Norway",
       "om": "Oman",
       "pk": "Pakistan",
       "pw": "Palau",
       "ps": "Palestinian Territory, Occupied",
       "pa": "Panama",
       "pg": "Papua New Guinea",
       "py": "Paraguay",
       "pe": "Peru",
       "ph": "Philippines",
       "pn": "Pitcairn",
       "pl": "Poland",
       "pt": "Portugal",
       "pr": "Puerto Rico",
       "qa": "Qatar",
       "re": "Reunion",
       "ro": "Romania",
       "ru": "Russian Federation",
       "rw": "Rwanda",
       "sh": "Saint Helena",
       "kn": "Saint Kitts and Nevis",
       "lc": "Saint Lucia",
       "pm": "Saint Pierre and Miquelon",
       "vc": "Saint Vincent and the Grenadines",
       "ws": "Samoa",
       "sm": "San Marino",
       "st": "Sao Tome and Principe",
       "sa": "Saudi Arabia",
       "sn": "Senegal",
       "sc": "Seychelles",
       "sl": "Sierra Leone",
       "sg": "Singapore",
       "sk": "Slovakia",
       "si": "Slovenia",
       "sb": "Solomon Islands",
       "so": "Somalia",
       "za": "South Africa",
       "gs": "South Georgia and the South Sandwich Islands",
       "es": "Spain",
       "lk": "Sri Lanka",
       "sd": "Sudan",
       "sr": "Suriname",
       "sj": "Svalbard and Jan Mayen",
       "sz": "Swaziland",
       "se": "Sweden",
       "ch": "Switzerland",
       "sy": "Syrian Arab Republic",
       "tw": "Taiwan",
       "tj": "Tajikistan",
       "tz": "Tanzania, United Republic of",
       "th": "Thailand",
       "tg": "Togo",
       "tk": "Tokelau",
       "to": "Tonga",
       "tt": "Trinidad and Tobago",
       "tn": "Tunisia",
       "tr": "Turkey",
       "tm": "Turkmenistan",
       "tc": "Turks and Caicos Islands",
       "tv": "Tuvalu",
       "ug": "Uganda",
       "ua": "Ukraine",
       "ae": "United Arab Emirates",
       "gb": "United Kingdom",
       "uk": "United Kingdom",
       "us": "United States",
       "um": "United States Minor Outlying Islands",
       "uy": "Uruguay",
       "uz": "Uzbekistan",
       "vu": "Vanuatu",
       "ve": "Venezuela",
       "vn": "Viet Nam",
       "vg": "Virgin Islands, British",
       "vi": "Virgin Islands, U.S.",
       "wf": "Wallis and Futuna",
       "eh": "Western Sahara",
       "ye": "Yemen",
       "yu": "Yugoslavia",
       "zm": "Zambia",
       "zw": "Zimbabwe",
    }
 
    country = None
    if tld in tld_countries:
        country = tld_countries[tld]
    return country



from Ganga.Utility.logging import getLogger
from Ganga.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm
from Ganga.Lib.LCG import Grid

logger = getLogger()


class LCGOutputDownloadTask(object):

    """
    Class for defining a data object for each output downloading task.
    """

    _attributes = ('gridObj', 'jobObj')

    def __init__(self, jobObj):
        self.jobObj = jobObj

    def __eq__(self, other):
        """
        download task comparison based on job's FQID.
        """
        if self.jobObj.getFQID('.') == other.jobObj.getFQID('.'):
            return True
        else:
            return False

    def __str__(self):
        """
        represents the task by the job object
        """
        return 'downloading task for job %s' % self.jobObj.getFQID('.')


class LCGOutputDownloadAlgorithm(Algorithm):

    """
    Class for implementing the logic of each downloading task.
    """

    def process(self, item):
        """
        downloads output of one LCG job 
        """

        pps_check = (True, None)

        job = item.jobObj

        # it is very likely that the job's downloading task has been
        # created and assigned in a previous monitoring loop
        # ignore such kind of cases
        if job.status in ['completing', 'completed', 'failed']:
            return True

        # it can also happen that the job was killed/removed by user between
        # the downloading task was created in queue and being taken by one of
        # the downloading thread. Ignore suck kind of cases
        if job.status in ['removed', 'killed']:
            return True

        job.updateStatus('completing')
        outw = job.getOutputWorkspace()

        pps_check = Grid.get_output(job.backend.id, outw.getPath(), job.backend.credential_requirements)

        if pps_check[0]:
            job.updateStatus('completed')
            job.backend.exitcode = 0
        else:
            job.updateStatus('failed')
            # update the backend's reason if the failure detected in the
            # Ganga's pps
            if pps_check[1] != 0:
                job.backend.reason = 'non-zero app. exit code: %s' % pps_check[
                    1]
                job.backend.exitcode = pps_check[1]

        # needs to update the master job's status to give an up-to-date status
        # of the whole job
        if job.master:
            job.master.updateMasterJobStatus()

        self.__appendResult__(job.getFQID('.'), True)

        return True


class LCGOutputDownloader(MTRunner):

    """
    Class for managing the LCG output downloading activities based on MTRunner.
    """

    def __init__(self, numThread=10):

        MTRunner.__init__(self, name='lcg_output_downloader', data=Data(
            collection=[]), algorithm=LCGOutputDownloadAlgorithm())

        self.keepAlive = True
        self.numThread = numThread

    def countAliveAgent(self):

        return self.__cnt_alive_threads__()

    def addTask(self, job):

        task = LCGOutputDownloadTask(job)

        logger.debug('add output downloading task: job %s' % job.getFQID('.'))

        self.addDataItem(task)

        return True



# File: GangaND280/__init__.py

from Ganga.Utility.Config import makeConfig
from Ganga.Utility.Config.Config import _after_bootstrap

if not _after_bootstrap:
   configND280 = makeConfig('ND280', 'Configuration parameters for the ND280 Ganga plugin')
   configND280.addOption('ND280DCacheDatasetCommandStr',
                         {'TRIUMF' : 'curl --compressed -sk --header "Accept: text/plain" https://user:password@nd280web.nd280.org/full/path/to/https/nd280data/%s'},
                         'How to get the list of files on the DCache server')
   configND280.addOption('ND280DCacheDatasetFileBasePath',
                         {'TRIUMF' : 'dcap://t2ksrm.nd280.org/full/path/to/dcap/nd280data/'},
                         'How to access files on the DCache sever')

def loadPlugins( config = {} ):

   import ND280Dataset
   import ND280Splitter
   import ND280Control
   import ND280Executable
   import ND280RecoValidation
   import ND280Skimmer
   import Highland
   import Tasks
   import ND280TPCGasInteractions
   import ND280Checkers

   return None



import copy
import os
import datetime
import inspect
import hashlib
import re
import os.path
import random
import glob
from Ganga.GPIDev.Base.Proxy import stripProxy, isType, getName
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Lib.File import FileUtils
from Ganga.GPIDev.Lib.Job.Job import Job
from Ganga.Utility.files import expandfilename
from Ganga.Core.exceptions import GangaFileError
from GangaDirac.Lib.Utilities.DiracUtilities import getDiracEnv, execute, GangaDiracError
import Ganga.Utility.Config
from Ganga.Runtime.GPIexport import exportToGPI
from Ganga.GPIDev.Credentials import require_credential
from GangaDirac.Lib.Credentials.DiracProxy import DiracProxy
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
config = getConfig('Configuration')
configDirac = getConfig('DIRAC')
logger = getLogger()
regex = re.compile('[*?\[\]]')

global stored_list_of_sites
stored_list_of_sites = []


class DiracFile(IGangaFile):

    """
    File stored on a DIRAC storage element
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                                     'localDir': SimpleItem(defvalue=None, copyable=1, typelist=['str', 'type(None)'],
                                                            doc='local dir where the file is stored, used from get and put methods'),
                                     'locations': SimpleItem(defvalue=[], copyable=1, typelist=['str'], sequence=1,
                                                             doc="list of SE locations where the outputfiles are uploaded"),
                                     'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0,
                                                              doc='wheather the output file should be compressed before sending somewhere'),
                                     'lfn': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                       doc='return the logical file name/set the logical file name to use if not '
                                                       'using wildcards in namePattern'),
                                     'remoteDir': SimpleItem(defvalue="", doc='remote directory where the LFN is to be placed within '
                                                             'the dirac base directory by the put method.'),
                                     'guid': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                        doc='return the GUID/set the GUID to use if not using wildcards in the namePattern.'),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], sequence=1, copyable=0,  # hidden=1,
                                                               typelist=['GangaDirac.Lib.Files.DiracFile'], doc="collected files from the wildcard namePattern"),
                                     'defaultSE': SimpleItem(defvalue='', copyable=1, doc="defaultSE where the file is to be accessed from or uploaded to"),
                                     'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure'),
                                     'credential_requirements': ComponentItem('CredentialRequirement', defvalue='DiracProxy'),
    })

    _env = None

    _category = 'gangafiles'
    _name = "DiracFile"
    _exportmethods = ["get", "getMetadata", "getReplicas", 'getSubFiles', 'remove',
                      "replicate", 'put', 'locations', 'location', 'accessURL',
                      '_updateRemoteURLs', 'hasMatchedFiles']
    _remoteURLs = {}
    _storedReplicas = {}
    _have_copied = False

    def __init__(self, namePattern='', localDir=None, lfn='', remoteDir=None, **kwds):
        """
        name is the name of the output file that has to be written ...
        """

        super(DiracFile, self).__init__()
        self.locations = []

        self._setLFNnamePattern(lfn, namePattern)

        if localDir is not None:
            self.localDir = localDir
        if remoteDir is not None:
            self.remoteDir = remoteDir

    def __setattr__(self, attr, value):
        """
        This is an overloaded setter method to make sure that we're auto-expanding the filenames of files which exist.
        In the case we're assigning any other attributes the value is simply passed through
        Args:
            attr (str): This is the name of the attribute which we're assigning
            value (unknown): This is the value being assigned.
        """
        actual_value = value
        if attr == "namePattern":
            actual_value = os.path.basename(value)
            this_dir = os.path.dirname(value)
            if this_dir:
                self.localDir = this_dir
        elif attr == 'localDir':
            if value:
                new_value = os.path.abspath(expandfilename(value))
                if os.path.exists(new_value):
                    actual_value = new_value

        super(DiracFile, self).__setattr__(attr, actual_value)

    def _attribute_filter__set__(self, name, value):

        if value != "" and value is not None:
            #   Do some checking of the filenames in a subprocess
            if name == 'lfn':
                self.namePattern = os.path.basename(value)
                if os.path.dirname(value):
                    self.remoteDir = os.path.dirname(value)
                return value

            elif name == 'remoteDir':
                if self.lfn != os.path.join(value, self.namePattern):
                    self.lfn = os.path.join(value, self.namePattern)

            elif name == 'namePattern':
                self.localDir = os.path.dirname(value)
                return os.path.basename(value)

            elif name == 'localDir':
                if value:
                    return expandfilename(value)
                else:
                    return value

        return value

    def locations(self):

        return self.locations

    def _setLFNnamePattern(self, lfn="", namePattern=""):

        if self.defaultSE != "":
            ## TODO REPLACE THIS WITH IN LIST OF VONAMES KNOWN
            # Check for /lhcb/some/path or /gridpp/some/path
            if namePattern.split(os.pathsep)[0] == self.defaultSE \
                or (len(namePattern) > 3 and namePattern[0:4].upper() == "LFN:"\
                    or len(namePattern.split(os.pathsep)) > 1 and namePattern.split(os.pathsep)[1] == self.defaultSE):
                # Check for LFN:/gridpp/some/path or others...
                lfn = namePattern
                namePattern = ""

        if lfn:
            if len(lfn) > 3 and lfn[0:4].upper() == "LFN:":
                lfn = lfn[4:]
        elif namePattern:
            if len(namePattern) > 3 and namePattern[0:4].upper() == 'LFN:':
                lfn = namePattern[4:]

        if lfn != "" and namePattern != "":
            self.lfn = lfn
            self.namePattern = namePattern

        elif lfn != "" and namePattern == "":
            self.lfn = lfn

        elif namePattern != "" and lfn == "":
            self.namePattern = namePattern

    def _attribute_filter__get__(self, name):

        # Attempt to spend too long loading un-needed objects into memory in
        # order to read job status
        if name is 'lfn':
            if self.lfn == "":
                logger.warning("Do NOT have an LFN, for file: %s" % self.namePattern)
                logger.warning("If file exists locally try first using the method put()")
            return object.__getattribute__(self, 'lfn')
        elif name in ['guid', 'locations']:
            if configDirac['DiracFileAutoGet']:
                if name is 'guid':
                    if self.guid is None or self.guid == '':
                        if self.lfn != "":
                            self.getMetadata()
                            return object.__getattribute__(self, 'guid')
                if name is 'locations':
                    if self.locations == []:
                        if self.lfn != "":
                            self.getMetadata()
                            return object.__getattribute__(self, 'locations')

        return object.__getattribute__(self, name)

    def __repr__(self):
        """Get the representation of the file."""

        return "DiracFile(namePattern='%s', lfn='%s', localDir='%s')" % (self.namePattern, self.lfn, self.localDir)

    def getSubFiles(self):
        """
        Returns the subfiles for this instance
        """
        if self.lfn == '':
            self.setLocation()
        return self.subfiles

    def dirac_line_processor(self, line, dirac_file, localPath):
        """
            Function to interperate the post processor lines.
            This returns False when everything went OK and True on an ERROR
        """
        logger.debug("Calling dirac_line_processor")
        tokens = line.strip().split(':::')
        logger.debug("dirac_line_processor: %s" % tokens)
        pattern = tokens[1].split('->')[0].split('&&')[0]
        name = tokens[1].split('->')[0].split('&&')[1]
        lfn = tokens[1].split('->')[1]
        guid = tokens[3]
        try:
            locations = eval(tokens[2])
        except Exception as err:
            logger.debug("line_process err: %s" % err)
            locations = tokens[2]

        if pattern == name:
            logger.debug("pattern == name")
            logger.error("Failed to parse outputfile data for file '%s'" % name)
            return True

        #   This is the case that multiple files were requested
        if pattern == dirac_file.namePattern:
            logger.debug("pattern == dirac_file.namePattern")
            d = DiracFile(namePattern=name, lfn=lfn)
            d.compressed = dirac_file.compressed
            d.guid = guid
            d.locations = locations
            d.localDir = localPath
            dirac_file.subfiles.append(d)
            #dirac_line_processor(line, d)
            return False

        #   This is the case that an individual file was requested
        elif name == dirac_file.namePattern:
            logger.debug("name == dirac_file.namePattern")
            if lfn == '###FAILED###':
                dirac_file.failureReason = tokens[2]
                logger.error("Failed to upload file '%s' to Dirac: %s" % (name, dirac_file.failureReason))
                return True
            dirac_file.lfn = lfn
            dirac_file.locations = locations
            dirac_file.guid = guid
            dirac_file.localDir = localPath
            return False

        else:
            logger.debug("False")
            return False

    def setLocation(self):
        """
        """

        logger.debug("DiracFile: setLocation")

        if not stripProxy(self).getJobObject():
            logger.error("No job assocaited with DiracFile: %s" % str(self))
            return

        job = self.getJobObject()
        postprocessLocationsPath = os.path.join(job.outputdir, getConfig('Output')['PostProcessLocationsFileName'])

        postprocesslocations = None

        try:
            postprocesslocations = open(postprocessLocationsPath, 'r')
            self.subfiles = []
            ## NB remember only do this once at it leaves the 'cursor' at the end of the file - rcurrie
            all_lines = postprocesslocations.readlines()
            logger.debug("lines:\n%s" % all_lines)
            for line in all_lines:
                logger.debug("This line: %s" % line)
                if line.startswith('DiracFile'):
                    if self.dirac_line_processor(line, self, os.path.dirname(postprocessLocationsPath)) and regex.search(self.namePattern) is None:
                        logger.error("Error processing line:\n%s\nAND: namePattern: %s is NOT matched" % (str(line), str(self.namePattern)))
                    else:
                        logger.debug("Parsed the Line")
                else:
                    logger.debug("Skipping the Line")

        except Exception as err:
            logger.warning("unexpected Error: %s" % str(err))
        finally:
            if postprocesslocations is not None:
                postprocesslocations.close()

    def _auto_remove(self):
        """
        Remove called when job is removed as long as config option allows
        """
        if self.lfn != '':
            self.remove()

    @require_credential
    def remove(self):
        """
        Remove this lfn and all replicas from DIRAC LFC/SEs
        """
        if self.lfn == "":
            raise GangaFileError('Can\'t remove a  file from DIRAC SE without an LFN.')
        logger.info('Removing file %s' % self.lfn)
        stdout = execute('removeFile("%s")' % self.lfn, cred_req=self.credential_requirements)

        self.lfn = ""
        self.locations = []
        self.guid = ''
        return True

    @require_credential
    def getMetadata(self):
        """
        Get Metadata associated with this files lfn. This method will also
        try to automatically set the files guid attribute.
        """

        if self.lfn == "":
            self._optionallyUploadLocalFile()

        # eval again here as datatime not included in dirac_ganga_server

        ret = execute('getMetadata("%s")' % self.lfn, cred_req=self.credential_requirements)

        if self.guid != ret.get('Successful',{}).get(self.lfn,{}).get('GUID',False):
            self.guid = ret['Successful'][self.lfn]['GUID']

        reps = self.getReplicas()
        ret['Successful'][self.lfn].update({'replicas': self.locations})

        return ret

    def _optionallyUploadLocalFile(self):
        """
        """

        if self.lfn != "":
            return

        if self.namePattern != "" and self.lfn == "":

            logger.info("I have a local DiracFile, however you're requesting it's location on the grid")
            logger.info("Shall I upload it to the grid before I continue?")
            decision = raw_input('[y] / n:')
            while not (decision.lower() in ['y', 'n'] or decision.lower() == ''):
                decision = raw_input('[y] / n:')

            if decision.lower() in ['y', '']:
                # upload namePattern to grid
                logger.debug("Uploading the file first")
                self.put()
            elif decision == 'n':
                logger.debug("Not uploading now")
                return
            else:
                # do Nothing
                logger.debug("Continuing without uploading file")

            if self.lfn == "":
                raise GangaFileError('Uploading of namePattern: %s failed' % self.namePattern)

        if self.namePattern == "" and self.lfn == "":
            raise GangaFileError('Cannot do anything if I don\'t have an lfn or a namePattern!')

        return

    @require_credential
    def getReplicas(self, forceRefresh=False):
        """
        Get the list of all SE where this file has a replica
        This relies on an internally stored list of replicas, (SE and  unless forceRefresh = True
        """

        if self.lfn == '':
            self._optionallyUploadLocalFile()
        if self.lfn == '':
            raise GangaFileError("Can't find replicas for file which has no LFN!")

        these_replicas = None

        if len(self.subfiles) != 0:

            allReplicas = []
            for i in self.subfiles:
                allReplicas.append(i.getReplicas())

            these_replicas = allReplicas

        else:
            # deep copy just before wer change it incase we're pointing to the
            # data stored in original from a copy
            if self._have_copied:
                self._storedReplicas = copy.deepcopy(self._storedReplicas)
            if (self._storedReplicas == {} and len(self.subfiles) == 0) or forceRefresh:

                try:
                    self._storedReplicas = execute('getReplicas("%s")' % self.lfn, cred_req=self.credential_requirements)
                except GangaDiracError as err:
                    logger.error("Couldn't find replicas for: %s" % str(self.lfn))
                    self._storedReplicas = {}
                    raise

                try:
                    self._storedReplicas = self._storedReplicas['Successful']
                except Exception as err:
                    logger.error("Unknown Error: %s from %s" % (str(err), self._storedReplicas))
                    raise

                logger.debug("getReplicas: %s" % str(self._storedReplicas))

                if self.lfn in self._storedReplicas:
                    self._updateRemoteURLs(self._storedReplicas)

                    these_replicas = [self._storedReplicas[self.lfn]]
                else:
                    these_replicas = {}
            elif self._storedReplicas != {}:
                these_replicas = [self._storedReplicas[self.lfn]]

        return these_replicas

    def _updateRemoteURLs(self, reps):
        """
        Internal function used for storing all replica information about this LFN at different sites
        """
        if len(self.subfiles) != 0:
            for i in self.subfiles:
                i._updateRemoteURLs(reps)
        else:
            if self.lfn not in reps:
                return
            if self.locations != reps[self.lfn].keys():
                self.locations = reps[self.lfn].keys()
            #logger.debug( "locations: %s" % str( self.locations ) )
            # deep copy just before wer change it incase we're pointing to the
            # data stored in original from a copy
            if self._have_copied:
                self._remoteURLs = copy.deepcopy(self._remoteURLs)
            for site in self.locations:
                #logger.debug( "site: %s" % str( site ) )
                self._remoteURLs[site] = reps[self.lfn][site]
                #logger.debug("Adding _remoteURLs[site]: %s" % str(self._remoteURLs[site]))

    def location(self):
        """
        Return a list of LFN locations for this DiracFile
        """
        if len(self.subfiles) == 0:
            if self.lfn == "":
                self._optionallyUploadLocalFile()
            else:
                return [self.lfn]
        else:
            # 1 LFN per DiracFile
            LFNS = []
            for this_file in self.subfiles:
                these_LFNs = this_file.location()
                for this_url in these_LFNs:
                    LFNs.append(this_url)
            return LFNs

    @require_credential
    def accessURL(self, thisSE=''):
        """
        Attempt to find an accessURL which corresponds to the specified SE. If no SE is specified then
        return a random one from all the replicas. 
        """
        from GangaDirac.Lib.Backends.DiracUtils import getAccessURLs
        lfns = []
        if len(self.subfiles) == 0:
            lfns.append(self.lfn)
        else:
            for i in self.subfiles:
                lfns.append(i.lfn)
        return getAccessURLs(lfns, thisSE)

    @require_credential
    def internalCopyTo(self, targetPath):
        """
        Retrieves locally the file matching this DiracFile object pattern.
        If localPath is specified
        Args:
            targetPath(str): The path the file should be placed at locally
        """

        to_location = targetPath

        if self.lfn == "":
            raise GangaFileError('Can\'t download a file without an LFN.')

        logger.info("Getting file %s" % self.lfn)
        stdout = execute('getFile("%s", destDir="%s")' % (self.lfn, to_location), cred_req=self.credential_requirements)

        if self.namePattern == "":
            name = os.path.basename(self.lfn)
            if self.compressed:
                name = name[:-3]
            self.namePattern = name

        if self.guid == "" or not self.locations:
            self.getMetadata()

        return True

    @require_credential
    def replicate(self, destSE, sourceSE=''):
        """
        Replicate an LFN to another SE

        Args:
            destSE (str): the SE to replicate the file to
            sourceSE (str): the se to use as a cource for the file
        """

        if not self.lfn:
            raise GangaFileError('Must supply an lfn to replicate')

        logger.info("Replicating file %s to %s" % (self.lfn, destSE))
        stdout = execute('replicateFile("%s", "%s", "%s")' % (self.lfn, destSE, sourceSE), cred_req=self.credential_requirements)

        if destSE not in self.locations:
            self.locations.append(destSE)

    def processWildcardMatches(self):
        if regex.search(self.namePattern) is not None:
            raise GangaFileError("No wildcards in inputfiles for DiracFile just yet. Dirac are exposing this in API soon.")

    @require_credential
    def put(self, lfn='', force=False, uploadSE="", replicate=False):
        """
        Try to upload file sequentially to storage elements defined in configDirac['allDiracSE'].
        File will be uploaded to the first SE that the upload command succeeds for.

        The file is uploaded to the SE described by the DiracFile.defaultSE attribute

        Alternatively, the user can specify an uploadSE which contains an SE
        which the file is to be uploaded to.

        If the user wants to replicate this file(s) across all SE then they should state replicate = True.

        Return value will be either the stdout from the dirac upload command if not
        using the wildcard characters '*?[]' in the namePattern.
        If the wildcard characters are used then the return value will be a list containing
        newly created DiracFile objects which were the result of glob-ing the wildcards.

        The objects in this list will have been uploaded or had their failureReason attribute populated if the
        upload failed.
        """

        if self.lfn != "" and force == False and lfn == '':
            logger.warning("Warning you're about to 'put' this DiracFile: %s on the grid as it already has an lfn: %s" % (self.namePattern, self.lfn))
            decision = raw_input('y / [n]:')
            while not (decision.lower() in ['y', 'n'] or decision.lower() == ''):
                decision = raw_input('y / [n]:')

            if decision.lower() == 'y':
                pass
            else:
                return

        if (lfn != '' and self.lfn != '') and force == False:
            logger.warning("Warning you're attempting to put this DiracFile: %s" % self.namePattern)
            logger.warning("It currently has an LFN associated with it: %s" % self.lfn)
            logger.warning("Do you want to continue and attempt to upload to: %s" % lfn)
            decision = raw_input('y / [n]:')
            while not (decision.lower() in ['y', 'n', '']):
                decision = raw_input('y / [n]:')

            if decision.lower() == 'y':
                pass
            else:
                return

        if lfn and os.path.basename(lfn) != self.namePattern:
            logger.warning("Changing namePattern from: '%s' to '%s' during put operation" % (self.namePattern, os.path.basename(lfn)))

        if lfn:
            self.lfn = lfn

        # looks like will only need this for the interactive uploading of jobs.
        # Also if any backend need dirac upload on client then when downloaded
        # this will upload then delete the file.

        if self.namePattern == "":
            if self.lfn != '':
                logger.warning("'Put'-ing a file with ONLY an existing LFN makes no sense!")
            raise GangaFileError('Can\'t upload a file without a local file name.')

        sourceDir = self.localDir
        if self.localDir is None:
            sourceDir = os.getcwd()
            # attached to a job, use the joboutputdir
            if self._parent != None and os.path.isdir(self.getJobObject().outputdir):
                sourceDir = self.getJobObject().outputdir

        if not os.path.isdir(sourceDir):
            raise GangaFileError('localDir attribute is not a valid dir, don\'t know from which dir to take the file')

        if regex.search(self.namePattern) is not None:
            if self.lfn != "":
                logger.warning("Cannot specify a single lfn for a wildcard namePattern")
                logger.warning("LFN will be generated automatically")
                self.lfn = ""

        if not self.remoteDir:
            try:
                job = self.getJobObject()
                lfn_folder = os.path.join("GangaUploadedFiles", "GangaJob_%s" % job.getFQID('.'))
            except AssertionError:
                t = datetime.datetime.now()
                this_date = t.strftime("%H.%M_%A_%d_%B_%Y")
                lfn_folder = os.path.join("GangaUploadedFiles", 'GangaFiles_%s' % this_date)
            self.lfn = os.path.join(DiracFile.diracLFNBase(), lfn_folder, self.namePattern)

        if self.remoteDir[:4] == 'LFN:':
            lfn_base = self.remoteDir[4:]
        else:
            lfn_base = self.remoteDir

        if uploadSE == "":
            if self.defaultSE != "":
                storage_elements = [self.defaultSE]
            else:
                if configDirac['allDiracSE']:
                    storage_elements = [random.choice(configDirac['allDiracSE'])]
                else:
                    raise GangaFileError("Can't upload a file without a valid defaultSE or storageSE, please provide one")
        elif isinstance(uploadSE, list):
            storage_elements = uploadSE
        else:
            storage_elements = [uploadSE]

        outputFiles = GangaList()
        for this_file in glob.glob(os.path.join(sourceDir, self.namePattern)):
            name = this_file

            if not os.path.exists(name):
                if not self.compressed:
                    raise GangaFileError('Cannot upload file. File "%s" must exist!' % name)
                name += '.gz'
                if not os.path.exists(name):
                    raise GangaFileError('File "%s" must exist!' % name)
            else:
                if self.compressed:
                    os.system('gzip -c %s > %s.gz' % (name, name))
                    name += '.gz'
                    if not os.path.exists(name):
                        raise GangaFileError('File "%s" must exist!' % name)

            if not lfn or not lfn.startswith(DiracFile.diracLFNBase()):
                if not lfn:
                    name = os.path.basename(name)
                lfn = os.path.join(DiracFile.diracLFNBase(), name)

            #lfn = os.path.join(os.path.dirname(self.lfn), this_file)

            d = DiracFile()
            d.namePattern = os.path.basename(name)
            d.compressed = self.compressed
            d.localDir = sourceDir
            stderr = ''
            stdout = ''
            logger.info('Uploading file \'%s\' to \'%s\' as \'%s\'' % (name, storage_elements[0], lfn))
            logger.debug('execute: uploadFile("%s", "%s", %s)' % (lfn, os.path.join(sourceDir, name), str([storage_elements[0]])))
            try:
                stdout = execute('uploadFile("%s", "%s", %s)' % (lfn, os.path.join(sourceDir, name), str([storage_elements[0]])), cred_req=self.credential_requirements)
            except GangaDiracError as err:
                logger.warning("Couldn't upload file '%s': \'%s\'" % (os.path.basename(name), err))
                failureReason = "Error in uploading file '%s' : '%s'" % (os.path.basename(name), err)
                if regex.search(self.namePattern) is not None:
                    d.failureReason = failureReason
                    outputFiles.append(d)
                    continue
                self.failureReason += '\n' + failureReason
                continue

            stdout_temp = stdout.get('Successful')

            if not stdout_temp:
                msg = "Couldn't upload file '%s': \'%s\'" % (os.path.basename(name), stdout)
                logger.warning(msg)
                if regex.search(self.namePattern) is not None:
                    d.failureReason = msg
                    outputFiles.append(d)
                    continue
                self.failureReason = msg
                continue
            else:
                lfn_out = stdout_temp[lfn]

            # when doing the two step upload delete the temp file
            if self.compressed or self._parent != None:
                os.remove(name)
            # need another eval as datetime needs to be included.
            guid = lfn_out.get('GUID', '')
            if regex.search(self.namePattern) is not None:
                d.lfn = lfn
                d.remoteDir = os.path.dirname(lfn)
                d.locations = lfn_out.get('allDiracSE', '')
                d.guid = guid
                outputFiles.append(d)
                continue
            else:
                self.lfn = lfn
                self.remoteDir = os.path.dirname(lfn)
                self.locations = lfn_out.get('allDiracSE', '')
                self.guid = guid

        if replicate == True:

            if len(outputFiles) == 1 or len(outputFiles) == 0:
                storage_elements.pop(0)
                for se in storage_elements:
                    self.replicate(se)
            else:
                storage_elements.pop(0)
                for this_file in outputFiles:
                    for se in storage_elements:
                        this_file.replicate(se)

        if len(outputFiles) > 0:
            return outputFiles
        else:
            outputFiles.append(self)
            return outputFiles

    def getWNScriptDownloadCommand(self, indent):

        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), 'downloadScript.py.template')

        download_script = FileUtils.loadScript(script_location, '')

        script = """\n
download_script='''\n###DOWNLOAD_SCRIPT###'''
import subprocess
dirac_env=###DIRAC_ENV###
subprocess.Popen('''python -c "import sys\nexec(sys.stdin.read())"''', shell=True, env=dirac_env, stdin=subprocess.PIPE).communicate(download_script)
"""
        script = '\n'.join([ str(indent+str(line)) for line in script.split('\n')])

        replace_dict = {'###DOWNLOAD_SCRIPT###' : download_script,
                        '###DIRAC_ENV###' : self._getDiracEnvStr(),
                        '###LFN###' : self.lfn}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def _getDiracEnvStr(self):
        diracEnv = str(getDiracEnv())
        return diracEnv

    def _WN_wildcard_script(self, namePattern, lfnBase, compressed):
        wildcard_str =  """
for f in glob.glob('###NAME_PATTERN###'):
    processes.append(uploadFile(os.path.basename(f), '###LFN_BASE###', ###COMPRESSED###, '###NAME_PATTERN###'))
"""
        wildcard_str = FileUtils.indentScript(wildcard_str, '###INDENT###')

        replace_dict = { '###NAME_PATTERN###' : namePattern,
                         '###LFN_BASE###' : lfnBase,
                         '###COMPRESSED###' : compressed }

        for k, v in replace_dict.iteritems():
            wildcard_str = wildcard_str.replace(str(k), str(v))

        return wildcard_str

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """

        script_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        script_location = os.path.join( script_path, 'uploadScript.py.template')

        upload_script = FileUtils.loadScript(script_location, '')

        WNscript_location = os.path.join( script_path, 'WNInjectTemplate.py.template' )
        script = FileUtils.loadScript(WNscript_location, '')

        if not self.remoteDir:
            try:
                job = self.getJobObject()
                lfn_folder = os.path.join("GangaUploadedFiles", "GangaJob_%s" % job.getFQID('.'))
            except AssertionError:
                t = datetime.datetime.now()
                this_date = t.strftime("%H.%M_%A_%d_%B_%Y")
                lfn_folder = os.path.join("GangaUploadedFiles", 'GangaFiles_%s' % this_date)
            self.lfn = os.path.join(DiracFile.diracLFNBase(), lfn_folder, self.namePattern)

        if self.remoteDir == '':
            self.remoteDir = DiracFile.diracLFNBase()

        if self.remoteDir[:4] == 'LFN:':
            lfn_base = self.remoteDir[4:]
        else:
            lfn_base = self.remoteDir

        for this_file in outputFiles:
            isCompressed = this_file.namePattern in patternsToZip

            if not regex.search(this_file.namePattern) is None:
                script += self._WN_wildcard_script(this_file.namePattern, lfn_base, str(isCompressed))
            else:
                script += '###INDENT###print("Uploading: %s as: %s")\n' % (this_file.namePattern, str(os.path.join(lfn_base, this_file.namePattern)))
                script += '###INDENT###processes.append(uploadFile("%s", "%s", %s))\n' % (this_file.namePattern, lfn_base, str(isCompressed))


        if stripProxy(self)._parent is not None and stripProxy(self).getJobObject() and getName(stripProxy(self).getJobObject().backend) != 'Dirac':
            script_env = self._getDiracEnvStr()
        else:
            script_env = str(None)

        script = '\n'.join([str('###INDENT###' + str(line)) for line in script.split('\n')])

        replace_dict = {'###UPLOAD_SCRIPT###' : upload_script,
                        '###STORAGE_ELEMENTS###' : str(configDirac['allDiracSE']),
                        '###INDENT###' : indent,
                        '###LOCATIONSFILE###' : postProcessLocationsFP,
                        '###DIRAC_ENV###' : script_env}

        for k, v in replace_dict.iteritems():
            script = script.replace(str(k), str(v))

        return script

    def hasMatchedFiles(self):

        if self.lfn != "" and self.namePattern != "":
            if self.namePattern == os.path.basename(self.lfn):
                return True
            else:
                logger.error("LFN doesn't match namePattern for file: %s" % str(self.namePattern))
                return False
        elif len(self.subfiles) > 0 and regex.search(self.namePattern) is not None:
            return True
        else:
            logger.error("Failed to Match file:\n%s" % str(self))
            return False

    @staticmethod
    def diracLFNBase():
        """
        Compute a sensible default LFN base name
        If ``DiracLFNBase`` has been defined, use that.
        Otherwise, construct one from the user name and the user VO
        """
        if configDirac['DiracLFNBase']:
            return configDirac['DiracLFNBase']
        return '/{0}/user/{1}/{2}'.format(configDirac['userVO'], config['user'][0], config['user'])

# add DiracFile objects to the configuration scope (i.e. it will be
# possible to write instatiate DiracFile() objects via config file)
Ganga.Utility.Config.config_scope['DiracFile'] = DiracFile

exportToGPI('GangaDirac', GangaList, 'Classes')




from Ganga.Core.exceptions import SplitterError
from Ganga.GPIDev.Adapters.ISplitter import SplittingError
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from Ganga.Utility.Config import getConfig
from Ganga.Utility.logging import getLogger
from GangaDirac.Lib.Utilities.DiracUtilities import execute, GangaDiracError
from Ganga.Core.GangaThread.WorkerThreads import getQueues
from GangaDirac.Lib.Files.DiracFile import DiracFile
from copy import deepcopy
import random
import time
import math

configDirac = getConfig('DIRAC')
logger = getLogger()

global_random = random

LFN_parallel_limit = 250
limit_divide_one = 1. / float(LFN_parallel_limit)

def wrapped_execute(command, expected_type):
    """
    A wrapper around execute to protect us from commands which had errors
    Args:
        command (str): This is the command to be exectuted against DIRAC
        expected_type (type): This is the type of the object which is returned from DIRAC
    """
    try:
        result = execute(command)
        assert isinstance(result, expected_type)
    except AssertionError:
        raise SplittingError("Output from DIRAC expected to be of type: '%s', we got the following: '%s'" % (expected_type, result))
    except GangaDiracError as err:
        raise SplittingError("Error from Dirac: %s" % err)
    return result


def find_random_site(original_SE_list, banned_SE):
    """
    Find a random element from a python list that isn't in a given banned list
    This is used for selecting a site element but it doesn't matter which is used

    Args:
        original_SE_list (list): This is a list of given 'SE'. The same SE may appear more than once!
        banned_SE (list): This is a list of SE which are 'banned' from being selected
    """
    input_list = deepcopy(original_SE_list)
    chosen_element = ""

    while chosen_element == "" and len(input_list) > 0:
        global global_random
        this_element = global_random.sample(input_list, 1)[0]
        if not this_element in banned_SE:
            chosen_element = this_element
            break
        else:
            input_list.remove(this_element)

    return chosen_element


def addToMapping(SE, site_to_SE_mapping):
    """
    This function is used for adding all of the known site for a given SE
    The output is stored in the dictionary site_to_SE_mapping
    Args:
        SE (str): For this SE we want to determine which sites we can access it
        site_to_SE_mapping (dict): We will add sites which can find this SE to this dict with the key (SE) and value (sites(list))
    """

    result = wrapped_execute('getSitesForSE("%s")' % str(SE), list)
    site_to_SE_mapping[SE] = result


def getLFNReplicas(allLFNs, index, allLFNData):
    """
    This method gets the location of all replicas for 'allLFNs' and stores the infomation in 'allLFNData'

    This is a 'static' method which is called multiple times with the same 'allLFNs' and 'allLFNData' and different index.
    e.g. This allows Dirac to determine the replicas for ~250LFN all at once rather than for ~40,000 all at once which risks timeouts and other errors

    Args:
        allLFNs (list): This is a list of all LFN which have replicas on the grid
        index (int): This is used to determine which slice of LFNs we want to look at
        allLFNData (dict): This is a dict where the replica information is to be stored temporarily
    """
    output = None

    global LFN_parallel_limit

    this_min = index * LFN_parallel_limit

    if (index + 1) * LFN_parallel_limit > len(allLFNs):
        this_max = len(allLFNs)
    else:
        this_max = (index + 1) * LFN_parallel_limit

    try:
        output = wrapped_execute('getReplicas(%s)' % str(allLFNs[this_min:this_max]), dict)
    except SplittingError:
        logger.error("Failed to Get Replica Info: [%s:%s] of %s" % (str(this_min), str(this_max), len(allLFNs)))
        raise

    import Ganga.Runtime.Repository_runtime
    Ganga.Runtime.Repository_runtime.updateLocksNow()

    allLFNData[index] = output

    logger.info("Got Replica Info: [%s:%s] of %s" % (str(this_min), str(this_max), len(allLFNs)))


def generate_site_selection(input_site, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping):
    """
    Return a set of sites which are a subset of the given input_site.
    The size of the returned set is determined by 'wanted_common_site'.
    'uniqueSE' governs if all elements in the returned set should be accessing uniqueSE or not.

    Args:
        input_site (list): list of input sites
        wanted_common_site (int): the size of the subset we want
        uniqueSE (bool): Should returned sites be able to share the same SE
        site_to_SE_mapping (dict): Dict which has sites as keys and SE as values
        SE_to_site_mapping (dict): Dict which has sites as values and SE as keys
    """

    req_sitez = set([])
    if len(input_site) > wanted_common_site:
        used_site = set([])
        for se in range(wanted_common_site):
            this_site = find_random_site(input_site, used_site)
            req_sitez.add(this_site)
            if uniqueSE:

                these_SE = SE_to_site_mapping[this_site]
                for this_SE in these_SE:

                    for site in site_to_SE_mapping[this_SE]:
                        used_site.add(site)
            else:
                used_site.add(this_site)
    else:
        for s in input_site:
            req_sitez.add(s)
    return req_sitez


def calculateSiteSEMapping(file_replicas, uniqueSE, site_to_SE_mapping, SE_to_site_mapping):
    """
    If uniqueSE:
        This constructs 2 dicts which allow for going between SE and sites based upon a key/value lookup.
        This generates this knowledge from looping through the LFN replicas at given sites and given SE.
    else:
        Don't construct a site<->SE mapping as it's not needed

    This returns a dict of which sites each LFN are accessible at

    Args:
        file_replicas (dict): This is the dictionary of LFN replicas with LFN as the key
        site_to_SE_mapping (dict): Dict which has sites as keys and SE as values
        SE_to_site_mapping (dict): Dict which has sites as values and SE as keys

    Returns:
        site_dict (dict): Dict of {'LFN':set([sites]), ...}
    """

    SE_dict = dict()
    maps_size = 0
    found = []

    logger.info("Calculating site<->SE Mapping")

    # First find the SE for each site
    for lfn, repz in file_replicas.iteritems():
        sitez = set([])
        if uniqueSE:
            for replica in repz:
                sitez.add(replica)
                if not replica in found:

                    getQueues()._monitoring_threadpool.add_function(addToMapping, (str(replica), site_to_SE_mapping))

                    maps_size = maps_size + 1
                    found.append(replica)

        SE_dict[lfn] = sitez

    # Doing this in parallel so wait for it to finish
    while len(site_to_SE_mapping) != maps_size:
        time.sleep(0.1)

    if uniqueSE:
        # Now calculate the 'inverse' dictionary of site for each SE
        for _SE, _sites in site_to_SE_mapping.iteritems():
            for site_i in _sites:
                if site_i not in SE_to_site_mapping:
                    SE_to_site_mapping[site_i] = set([])
                if _SE not in SE_to_site_mapping[site_i]:
                    SE_to_site_mapping[site_i].add(_SE)

    # These can be used to select the site which know of a given SE
    # Or vice versa

    # Now lets generate a dictionary of some chosen site vs LFN to use in
    # constructing subsets
    site_dict = {}
    for _lfn, sites in SE_dict.iteritems():
        site_dict[_lfn] = set([])
        for _site in sites:
            for _SE in site_to_SE_mapping[_site]:
                site_dict[_lfn].add(_SE)

    return site_dict


def lookUpLFNReplicas(inputs, ignoremissing):
    """
    This method launches several worker threads to collect the replica information for all LFNs which are given as inputs and stores this in allLFNData
    Args:
        inputs (list): This is a list of input DiracFile which are 
    Returns:
        allLFNs (list): List of all of the LFNs in the inputs
        LFNdict (dict): dict of LFN to DiracFiles
    """
    allLFNData = {}
    # Build a useful dictionary and list
    allLFNs = [_lfn.lfn for _lfn in inputs]
    LFNdict = dict.fromkeys(allLFNs)
    for _lfn in inputs:
        LFNdict[_lfn.lfn] = _lfn

    # Request the replicas for all LFN 'LFN_parallel_limit' at a time to not overload the
    # server and give some feedback as this is going on
    global limit_divide_one
    for i in range(int(math.ceil(float(len(allLFNs)) * limit_divide_one))):

        getQueues()._monitoring_threadpool.add_function(getLFNReplicas, (allLFNs, i, allLFNData))

    while len(allLFNData) != int(math.ceil(float(len(allLFNs)) * limit_divide_one)):
        time.sleep(1.)
        # This can take a while so lets protect any repo locks
        import Ganga.Runtime.Repository_runtime
        Ganga.Runtime.Repository_runtime.updateLocksNow()

    bad_lfns = []

    # Sort this information and store is in the relevant Ganga objects
    badLFNCheck(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData)

    # Check if we have any bad lfns
    if bad_lfns and ignoremissing is False:
        logger.error("Errors found getting LFNs:\n%s" % str(bad_lfns))
        raise SplittingError("Error trying to split dataset with invalid LFN and ignoremissing = False")

    return allLFNs, LFNdict, bad_lfns


def badLFNCheck(bad_lfns, allLFNs, LFNdict, ignoremissing, allLFNData):
    """
    Method to re-sort the LFN replica data and check for bad LFNs

    Args:
        bad_lfns (list): This is the list which will contain LFNs which have no replicas
        allLFNs (list): List of all of the LFNs in the inputs which have accessible replicas
        LFNdict (dict): dict of LFN to DiracFiles
        ignoremissing (bool): Check if we have any bad lfns
        allLFNData (dict): All LFN replica data
    """

    # FIXME here to keep the repo settings as they were before we changed the
    # flush count
    original_write_perm = {}

    global LFN_parallel_limit
    global limit_divide_one

    for i in range(int(math.ceil(float(len(allLFNs)) * limit_divide_one))):
        output = allLFNData.get(i)

        if output is None:
            msg = "Error getting Replica information from Dirac: [%s,%s]" % ( str(i * LFN_parallel_limit), str((i + 1) * LFN_parallel_limit))
            raise SplittingError(msg)

        # Identify files which have Failed to be found by DIRAC
        results = output
        values = results.get('Successful')

        upper_limit = (i+1)*LFN_parallel_limit
        if upper_limit > len(allLFNs):
            upper_limit = len(allLFNs)

        #logger.debug("Updating LFN Physical Locations: [%s:%s] of %s" % (str(i * LFN_parallel_limit), str(upper_limit), str(len(allLFNs))))

        for this_lfn in values.keys():
            #logger.debug("LFN: %s" % str(this_lfn))
            this_dict = {}
            this_dict[this_lfn] = values.get(this_lfn)

            if this_lfn in LFNdict:
                #logger.debug("Updating RemoteURLs")
                LFNdict[this_lfn]._updateRemoteURLs(this_dict)
                #logger.debug("This_dict: %s" % str(this_dict))
            else:
                logger.error("Error updating remoteURLs for: %s" % str(this_lfn))

            # If we find NO replicas then also mark this as bad!
            if this_dict[this_lfn].keys() == []:
                bad_lfns.append(this_lfn)

        for this_lfn in bad_lfns:
            logger.warning("LFN: %s was either unknown to DIRAC or unavailable, Ganga is ignoring it!" % str(this_lfn))
            if this_lfn in LFNdict:
                del LFNdict[this_lfn]
            if this_lfn in allLFNs:
                allLFNs.remove(this_lfn)


# Actually Do the work of the splitting

def OfflineGangaDiracSplitter(_inputs, filesPerJob, maxFiles, ignoremissing):
    """
    Generator that yields a datasets for dirac split jobs

    Args:
        _inputs (list): This is a list of input DiracFile objects
        filesPerJob (int): Max files per jobs as defined by splitter
        maxFiles (int): This is the max number of files per subset(subjob)
        ignoremissing (bool): Should we ignore missing LFN

    Yields:
        dataset (list): A list of LFNs for each subset(subjob)
    """

    if maxFiles is not None and maxFiles > 0:
        inputs = _inputs[:maxFiles]
    else:
        inputs = _inputs

    # First FIND ALL LFN REPLICAS AND SE<->SITE MAPPINGS AND STORE THIS IN MEMORY
    # THIS IS DONE IN PARALLEL TO AVOID OVERLOADING DIRAC WITH THOUSANDS OF
    # REQUESTS AT ONCE ON ONE CONNECTION

    wanted_common_site = configDirac['OfflineSplitterMaxCommonSites']
    uniqueSE = configDirac['OfflineSplitterUniqueSE']

    if inputs is None:
        raise SplittingError("Cannot Split Job as the inputdata appears to be None!")

    if len(inputs.getLFNs()) != len(inputs.files):
        raise SplittingError("Error trying to split dataset using DIRAC backend with non-DiracFile in the inputdata")

    file_replicas = {}

    logger.info("Requesting LFN replica info")

    # Perform a lookup of where LFNs are all stored
    allLFNs, LFNdict, bad_lfns = lookUpLFNReplicas(inputs, ignoremissing)

    # This finds all replicas for all LFNs...
    # This will probably struggle for LFNs which don't exist
    # Bad LFN should have been removed by this point however
    all_lfns = [LFNdict[this_lfn].locations for this_lfn in LFNdict if this_lfn not in bad_lfns]

    logger.info("Got all good replicas")

    for this_input in inputs:
        if this_input.lfn not in bad_lfns:
            file_replicas[this_input.lfn] = this_input.locations

    logger.info("found all replicas")


    site_to_SE_mapping = {}
    SE_to_site_mapping = {}

    allSubSets = []

    # Now lets generate a dictionary of some chosen site vs LFN to use in
    # constructing subsets
    site_dict = calculateSiteSEMapping(file_replicas, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)


    allChosenSets = {}
    # Now select a set of site to use as a seed for constructing a subset of
    # LFN
    for lfn in site_dict.keys():
        allChosenSets[lfn] = generate_site_selection(site_dict[lfn], wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)

    logger.debug("Found all SE in use")


    # BELOW IS WHERE THE ACTUAL SPLITTING IS DONE

    logger.info("Calculating best data subsets")

    allSubSets = performSplitting(site_dict, filesPerJob, allChosenSets, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)

    avg = 0.
    for this_set in allSubSets:
        avg += float(len(this_set))
    avg /= float(len(allSubSets))

    logger.info("Average Subset size is: %s" % (str(avg)))

    # FINISHED SPLITTING CHECK!!!
    check_count = 0
    for i in allSubSets:
        check_count = check_count + len(i)

    if check_count != len(inputs) - len(bad_lfns):
        logger.error("SERIOUS SPLITTING ERROR!!!!!")
        raise SplitterError("Files Missing after Splitting!")
    else:
        logger.info("File count checked! Ready to Submit")

    # RETURN THE RESULT

    logger.info("Created %s subsets" % len(allSubSets))

    for dataset in allSubSets:
        yield dataset

def performSplitting(site_dict, filesPerJob, allChosenSets, wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping):
    """
    This is the main method which loops through the LFNs and creates subsets which are returned a list of list of LFNs

    Args:
        site_dict (dict): This is a dict with LFNs as keys and sites for each LFN as value
        filesPerJob (int): Max files per jobs as defined by splitter
        allChosenSets (dict): A dict with LFNs as keys and a sub-set of sites where each LFN is replicated
        wanted_common_site (int): Number of sites which we want to have in common for each LFN

        uniqueSE (bool): Should we check to make sure sites don't share an SE
        site_to_SE_mapping (dict): Dict which has sites as keys and SE as values
        SE_to_site_mapping (dict): Dict which has sites as values and SE as keys

    Returns:
        allSubSets (list): Return a list of subsets each subset being a list of LFNs
    """

    good_fraction = configDirac['OfflineSplitterFraction']
    iterative_limit = configDirac['OfflineSplitterLimit']

    allSubSets = []

    iterations = 0
    # Loop over all LFNs
    while len(site_dict.keys()) > 0:

        # LFN left to be used
        # NB: Can't modify this list and iterate over it directly in python
        LFN_instances = site_dict.keys()
        # Already used LFN
        chosen_lfns = set()

        for iterating_LFN in LFN_instances:

            # If this has previously been selected lets ignore it and move on
            if iterating_LFN in chosen_lfns:
                continue

            # Use this seed to try and construct a subset
            req_sitez = allChosenSets[iterating_LFN]
            _this_subset = []

            #logger.debug("find common LFN for: " + str(allChosenSets[iterating_LFN]))

            # Construct subset
            # Starting with i, populate subset with LFNs which have an
            # overlap of at least 2 SE
            for this_LFN in LFN_instances:
                if this_LFN in chosen_lfns:
                    continue
                if req_sitez.issubset(site_dict[this_LFN]):
                    if len(_this_subset) >= filesPerJob:
                        break
                    _this_subset.append(this_LFN)

            limit = int(math.floor(float(filesPerJob) * good_fraction))

            #logger.debug("Size limit: %s" % str(limit))

            # If subset is too small throw it away
            if len(_this_subset) < limit:
                #logger.debug("%s < %s" % (str(len(_this_subset)), str(limit)))
                allChosenSets[iterating_LFN] = generate_site_selection(site_dict[iterating_LFN], wanted_common_site, uniqueSE, site_to_SE_mapping, SE_to_site_mapping)
                continue
            else:
                logger.debug("found common LFN for: " + str(allChosenSets[iterating_LFN]))
                logger.debug("%s > %s" % (str(len(_this_subset)), str(limit)))
                # else Dataset was large enough to be considered useful
                logger.debug("Generating Dataset of size: %s" % str(len(_this_subset)))
                ## Construct DiracFile here as we want to keep the above combination
                allSubSets.append([DiracFile(lfn=str(this_LFN)) for this_LFN in _this_subset])

                for lfn in _this_subset:
                    site_dict.pop(lfn)
                    allChosenSets.pop(lfn)
                    chosen_lfns.add(lfn)

        # Lets keep track of how many times we've tried this
        iterations = iterations + 1

        # Can take a while so lets not let threads become un-locked
        import Ganga.Runtime.Repository_runtime
        Ganga.Runtime.Repository_runtime.updateLocksNow()

        # If on final run, will exit loop after this so lets try and cleanup
        if iterations >= iterative_limit:

            if good_fraction < 0.5:
                good_fraction = good_fraction * 0.75
                iterations = 0
            elif wanted_common_site > 1:
                logger.debug("Reducing Common Site Size")
                wanted_common_site = wanted_common_site - 1
                iterations = 0
                good_fraction = 0.75
            else:
                good_fraction = good_fraction * 0.75

            logger.debug("good_fraction: %s" % str(good_fraction))

    return allSubSets



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PDF.py,v 1.1 2009-05-10 16:42:03 karl Exp $
###############################################################################
# File: PDF.py
# Author: K. Harrison
# Created: 070122

"""Module containing class for PDF application"""

__author__  = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__    = "9 May 2009"
__version__ = "1.0"

import os
import shutil

from Ganga.Core import ApplicationConfigurationError
from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Schema import ComponentItem, Schema, SimpleItem, Version
from Ganga.Utility import logging
from Ganga.Utility.files import fullpath

logger = logging.getLogger()

class PDF( IApplication ):

   _schema = Schema( Version( 1, 0 ), {
      "version" : SimpleItem( defvalue = "1.0", doc = \
         "String giving version number" ),
      "softwareDir" : SimpleItem( defvalue = "", doc = \
         "\n".join( [ "Directory containing application software", \
         "Used only if VO_CAMONT_SW_DIR is undefined" ] ) ),
      "outDir" : SimpleItem( defvalue = "", doc = \
         "\n".join( [ "Directory for output files", \
         "If null, use top-level directory of input documents" ] ) ),
      "docList" : SimpleItem( defvalue = "", doc = \
         "Path to file containing list of input documents" ),
      "maxDoc" : SimpleItem( defvalue = 0, doc = \
         "Maximum number of documents to process, "\
         + "no maximum for negative value" ),
      } )

   _category = 'applications'
   _name = 'PDF'

   def __init__( self ):
      super( PDF, self ).__init__()
        
   def master_configure( self ):

      listFile = fullpath( self.docList )
      if not os.path.isfile( listFile ):
         raise ApplicationConfigurationError\
            ( None, "docList file = '%s' not found" % self.docList )

      docFile = open( listFile )
      docDir = docFile.readline().strip().rstrip( os.sep )
      inputList = []

      if docDir: 
         docDir = fullpath( docDir )
         inputList = docFile.readlines()
         docFile.close()
         outDir = fullpath( self.outDir )
         if not outDir:
            outDir = docDir
         tagFile = ""
      else:
         outDir = fullpath( self.outDir )
         tagFile = listFile

      if not outDir:
         raise ApplicationConfigurationError\
            ( None, "Output directory not defined" )

      if not os.path.isdir( outDir ):
         os.makedirs( outDir )

      docList = []
      for input in inputList:
         url = input.split()[ 0 ]
         relativePath = input[ len( url ) : ].strip().strip( os.sep )
         docList.append( ( url, relativePath ) )

         if len( docList ) >= self.maxDoc:
            if abs( self.maxDoc ) == self.maxDoc:
               break

      if self.softwareDir:
        softwareDir = fullpath( self.softwareDir )
      else:
        softwareDir = ""

      appDict = \
         {
         "version" : self.version,
         "docDir" : docDir,
         "outDir" : outDir,
         "docList" : docList,
         "tagFile" : tagFile,
         "softwareDir" : softwareDir
         }

      return( False, appDict )

   def configure( self, master_appconfig ):    

      return ( False, master_appconfig )

   def postprocess( self ):
      job = self._getParent()
      try:
         outputdata = job.outputdata
         datatype = outputdata._name
      except AttributeError:
         outputdata = None
         datatype = None

      if datatype:
         outputdata.fill()

      return None



from __future__ import print_function
from Ganga.Utility.Config import getConfig, setSessionValuesFromFiles

if __name__ == "__main__":

    import Ganga.Utility.logging

    Ganga.Utility.logging.config['Ganga.Utility.Config'] = 'DEBUG'
    Ganga.Utility.logging.bootstrap()

    print('Basic Test')
    import sys

    c1 = getConfig('C1')
    c1.setDefaultOptions({'a': 1, 'aa': 'xx'})

    c2 = getConfig('C2')
    c2['b'] = 2
    c2['bb'] = 'yy'

    print('path=', sys.path[0])

    setSessionValuesFromFiles([sys.path[0] + '/ConfigTest.ini'])

    print("C1")
    print(c1.getEffectiveOptions())
    print()
    print("C2")
    print(c2.getEffectiveOptions())

    assert(c1['a'] == 3)
    assert(c1['aa'] == 'xx')

    assert(c2['b'] == 4)
    assert(c2['c'] == 'x')
    assert(c2['d'] == c2['c'])

    print('Basic Test OK')
    print()
    print('Hierarchical Test')

    c = getConfig('hierarchy')

    setSessionValuesFromFiles([sys.path[0] + '/A.ini', sys.path[0] + '/B.ini'])

    print(c.getEffectiveOptions())

    assert(c['MY_PATH'] == 'a2:a')
    assert(c['YOURPATH'] == 'b2')
    assert(c['PATH'] == 'c2')
    assert(c['MY2_PATH'] == 'd2')

    print('Hierarchical Test OK')



#
# CRAB Application
# 
# 08/06/10 @ ubeda
#
import subprocess

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Schema import *
from Ganga.Utility.logging import getLogger

from GangaCMS.Lib.ConfParams import *
from GangaCMS.Lib.CRABTools.CRABServer import *

import Ganga.Utility.Config

logger = getLogger()

class CRABApp(IApplication):

  comments=[]
  comments.append('crab.cfg file either created by GangaCMS or added by user.')
  comments.append('Workdir.')

  schemadic = {}
  schemadic['cfgfile']        = SimpleItem(defvalue=None,    typelist=['type(None)','str'], doc=comments[0], copiable=0) 
  schemadic['workdir']        = SimpleItem(defvalue=None,    typelist=['type(None)','str'], doc=comments[1], copiable=0) 

  # is_prepared is needed for GangaRobot on Ganga 5.7.0 and later.
  schemadic['is_prepared']    = SharedItem(defvalue=None,
                                           strict_sequence=0,
                                           visitable=1,
                                           copyable=1,
                                           typelist=['type(None)','bool','str'],
                                           protected=0,
                                           doc='Location of shared resources. Presence of this attribute implies the application has been prepared.')

  _schema = Schema(Version(1,0), schemadic)
  _category = 'applications'
  _name = 'CRABApp' 

  def __init__(self):
#    self.server = CRABServer()
    super(CRABApp,self).__init__()

  def writeCRABFile(self,job,cfg_file):

    file = open(cfg_file,'w')     

    job.inputdata.ui_working_dir = job.outputdir
                       
    for params in [CMSSW(),CRAB(),GRID(),USER()]:

      section = params.__class__.__name__

      config = Ganga.Utility.Config.getConfig('%s_CFG'%(section))
      file.write('['+section+']\n\n')

      for k in params.schemadic.keys():

        # We get the parametef from the config. If it is not NULL,
        # we use that, otherwise, we try to take it from inputdata.
        attr = config[k]
        if attr == None:
          attr = getattr(job.inputdata,k)

        if attr != None:  
          file.write('%s=%s\n'%(k,attr))                  
      file.write('\n')

    file.close()

  def master_configure(self):

    #Get job containing this CRABApp 
    job = self.getJobObject()

    #File where crab.cfg is going to be written
    cfg_file = '%scrab.cfg'%(job.inputdir)
    job.application.writeCRABFile(job,cfg_file)
    job.application.cfgfile = cfg_file

    job.application.workdir = job.inputdata.ui_working_dir

    server = CRABServer()
    server.create(job)
#    self.server.create(job)

    return (1,None)

  def configure(self,masterappconfig):
    return (None,None)



"""Threaded submitter IAction implementation.

The ThreadedSubmitter class provides a generic submitter implementation.
 
"""

from GangaRobot.Lib.Core.CoreSubmitter import CoreSubmitter
from Ganga.Utility.logging import getLogger
from Ganga.GPI import load
from Ganga.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm
from Ganga.GPIDev.Base.Proxy import stripProxy

logger = getLogger()


class ThreadedSubmitter(CoreSubmitter):

    """Threaded submitter IAction implementation.
    
    A submitter which submits exported jobs in a configurable number of threads. See handlesubmit() for details.
    
    """

    def handlesubmit(self, jobids, runid):
        """Submits exported jobs as identified by a list of path patterns.
        
        Keyword arguments:
        jobids -- A list of the submitted job ids.
        runid -- A UTC ID string which identifies the run.
        
        """
        # get configuration properties
        patterns = self.getoption('CoreSubmitter_Patterns')
        logger.info("Searching for job files matching patterns %s.", patterns)
        matches = self._getmatches(patterns)
        logger.info("Found %d matching job files.", len(matches))
        runner = MTRunner(
            'ThreadedSubmitterMTRunner',
            algorithm=ThreadedSubmitterAlgorithm(), 
            data=Data([(m,jobids) for m in matches]), 
            numThread=int(self.getoption('ThreadedSubmitter_numThreads'))
        )
        runner.start()
        runner.join()


class ThreadedSubmitterAlgorithm(Algorithm):
    def process(self, item):
        (match,jobids) = item
        jobs = load(match)
        logger.info("Loaded %d jobs from '%s'.", len(jobs), match)
        for j in jobs:
            stripProxy(j.application).is_prepared = True
            j.submit()
            jobids.append(j.id)




##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: $
###############################################################################
# A DQ2 dataset superclass, with AMI connection and search capability

from GangaAtlas.Lib.ATLASDataset.DQ2Dataset import DQ2Dataset, listDatasets
from Ganga.GPIDev.Schema.Schema import SimpleItem
from Ganga.GPIDev.Schema.Schema import FileItem
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig, ConfigError

import os

logger = getLogger()
config = getConfig('AMIDataset')

try:
    from pyAMI.pyAMI import AMI as AMIClient
    from pyAMI.pyAMIEndPoint import pyAMIEndPoint

    #Reading from CERN replica
    pyAMIEndPoint.setType("replica")
    amiclient = AMIClient()

except ImportError:
    logger.warning("AMI not properly set up. Set athena to access AMI from this ganga session.")
    pass

def resolve_dataset_name(name = ''):
    result = listDatasets(name + '*')
    status = False
    for dset in result:
        if dset[0] == name + '/':
            status = True
    return status

def get_metadata(dataset = '', file_name = ''):
    
    argument = []
    argument.append("GetDatasetInfo")
    if dataset:
        argument.append("logicalDatasetName=" + dataset)
        limit="0,%d" %config['MaxNumOfDatasets']
    else:
        argument.append("logicalFileName=" + file_name)
        limit="0,%d" %config['MaxNumOfFiles']

    argument.append("limit=" + limit)
    #Dictionary which contain all metadata info about a dataset or file
    metadata = {}       

    try:
        result =  amiclient.execute(argument)
        resultDict = result.getDict()
        resultByRow = resultDict['Element_Info']
        for rows, ids in resultByRow.iteritems():
            for metaid, id in ids.iteritems():
                metadata[str(metaid)] = str(id)

    except Exception as msg:
        logger.warning("Couldn't get metadata from AMI due to %s" % msg)

    return metadata     


def get_file_metadata(dataset='', all=False, numevtsperfile = 0):
    
    argument = []
    argument.append("ListFiles")
    argument.append("logicalDatasetName=" + dataset)
    limit="0,%d" %config['MaxNumOfFiles']
    argument.append("limit=" + limit)
    
    #Metatdata from all the files from a dataset
    info = []

    try:
        result =  amiclient.execute(argument)
        resultDict = result.getDict()
        resultByRow = resultDict['Element_Info']
        for rows, ids in resultByRow.iteritems():
            if all:
                file_metadata = get_metadata(file_name = ids['LFN'])
            else:
                file_metadata = {}
            for id,val in ids.iteritems():
                attr = str(id).lower()
                if ( (attr == 'filesize' or attr == 'events') and len(val) > 0):
                    val = int(val)
                file_metadata[attr]  = val 
            
            nevents = file_metadata.setdefault('events', 0)
            info.append(file_metadata)
    
    except Exception as msg:
        logger.warning("Couldn't get file metadata from AMI due to %s" % msg)
    
    content = {}
    for i in info:
        guid = str(i.pop('fileguid'))
        content[guid] = i
    
    return content 


class AMIDataset(DQ2Dataset):
    '''ATLAS DDM Dataset With AMI Connection'''

    _category = 'datasets'
    _name = 'AMIDataset'
            
    _schema = DQ2Dataset._schema.inherit_copy()
    _schema.datadict['logicalDatasetName'] = SimpleItem(defvalue = '', doc = '')
    _schema.datadict['project'] = SimpleItem(defvalue = 'Atlas_Production', doc = '')
    _schema.datadict['processingStep'] = SimpleItem(defvalue = 'Atlas_Production', doc = '')
    _schema.datadict['amiStatus'] = SimpleItem(defvalue = 'VALID', doc = '')
    _schema.datadict['entity'] = SimpleItem(defvalue = 'dataset', doc = '')
    _schema.datadict['metadata'] = SimpleItem(defvalue = {}, doc="Metadata" )
    _schema.datadict['provenance'] = SimpleItem(defvalue=[], typelist=['str'], sequence=1, doc='Dataset provenance chain')
    _schema.datadict['goodRunListXML'] = FileItem(doc = 'GoodRunList XML file to search on')


    _exportmethods = ['search','fill_provenance', 'get_datasets_metadata', 'get_files_metadata', 'get_contents']


    def __init__(self):
        super( AMIDataset, self ).__init__()

    def fill_provenance(self,extraargs = []):

        dataType=""
        if(len(extraargs)>1 ):
            dataType= extraargs[1]
                                
        self.provenance = []

        for d in self.dataset:

            logger.info("Filling provenance info for dataset %s...", d )

            prov = []
            self.provenance.append(prov)
            
            ds = d[:-1]

            argument=[]
            argument.append("ListDatasetProvenance")
            argument.append("logicalDatasetName="+ds)
            argument.append('-output=xml')

            result= amiclient.execute(argument)

            dom = result.getAMIdom()
            graph = dom.getElementsByTagName('graph')
            nFound = 0
            dictOfLists={}
            for line in graph:
                nodes = line.getElementsByTagName('node')
                for node in nodes:
                    level=int(node.attributes['level'].value)
                    dataset = node.attributes['name'].value
                    if (len(dataType)>0)and(dataset.find(dataType)>=0):
                        # print only selected dataType
                        levelList=dictOfLists.get(level,[])
                        levelList.append(dataset)
                        dictOfLists[level] = levelList 
                        nFound=nFound+1
                    elif (len(dataType)== 0):
                        #print everything 
                        levelList=dictOfLists.get(level,[])
                        levelList.append(dataset)    
                        dictOfLists[level] = levelList
                        #print level,dictOfLists[level]
                        nFound=nFound+1
            if(nFound==0)and (len(dataType)>0):
                logger.warning( "No datasets found of type",dataType)
            else:
                keys = dictOfLists.keys()
            
                keys.sort()

                for key in keys:
                    datasetList=dictOfLists.get(key)
                    datasetList.sort()                    
                    #print "generation =",key
                    #for dataset in datasetList:
                    #    print " ",dataset 
                    for dataset in datasetList:
                        prov.append("%s/" % dataset.strip())

    def search(self, pattern='', maxresults = config['MaxNumOfDatasets'], extraargs = []):
        
        argument=[]
        dsetList = []
        
        if self.goodRunListXML.name != '':

            # open the GRL
            if os.path.exists( self.goodRunListXML.name ):
                logger.info("Good Run List '%s' file selected" % self.goodRunListXML.name)
                grl_text = open( self.goodRunListXML.name ).read()
            else:
                logger.error('Could not read Good Run List XML file')
                return []
            
            argument=[]
            argument.append("GetGoodDatasetList")
            argument.append("prodStep=merge")
            #argument.append("dataType=%s" % self.dataType)
            argument.append("goodRunList=%s" % grl_text)
            argument.append("logicalDatasetName=%s" % self.logicalDatasetName)
                   
        elif self.logicalDatasetName:
            pattern=self.logicalDatasetName
            
            pattern = pattern.replace("/","")    
            pattern = pattern.replace('*','%')
            limit="0,%d" %config['MaxNumOfDatasets']

            argument.append("SearchQuery")
            argument.append("entity=" + self.entity)

            argument.append("glite=SELECT logicalDatasetName WHERE amiStatus='" + self.amiStatus +"' AND logicalDatasetName like '" + pattern +"' LIMIT "+limit)

            argument.append("project=" + self.project)
            argument.append("processingStep=" + self.processingStep)
            argument.append("mode=defaultField")
            argument.extend(extraargs)

        else:
            logger.error("AMI search not set up correctly. No datasetname or good runs list supplied.")
            return []
        
        try:
            result = amiclient.execute(argument)
            if argument[0] == "GetGoodDatasetList":
                # GRL has different output
                res_text = result.output()
                dsetList = []
                for ln in res_text.split('\n'):
                    if ln.find("logicalDatasetName") != -1:
                        # add to the dataset list - check container...
                        if resolve_dataset_name(ln.split('=')[1].strip()):
                                dsetList.append(ln.split('=')[1].strip() + "/")
                        else:
                                dsetList.append(ln.split('=')[1].strip())
                                
            else:
                resultDict= result.getDict()
                resultByRow = resultDict['Element_Info']
                for row, vals in resultByRow.iteritems():
                    dsName = str(vals['logicalDatasetName'])
                    # check with DQ2 since AMI doesn't store /
                    if resolve_dataset_name(dsName):
                        dsName += '/'
                    dsetList.append(dsName)
                    
        except Exception as msg:
            logger.error( msg )

        return dsetList

    def get_datasets_metadata(self):
        datasets = self.search()
        metadata = []

        for dataset in datasets:
            dataset = dataset.replace("/","")    
            tmp = get_metadata(dataset = dataset)
            metadata.append(tmp)

        return metadata
    
    def get_files_metadata(self, all=False):
        datasets = self.search()
        metadata = {}

        for dataset in datasets:
            dataset = dataset.replace("/","")    
            job = self._getParent()
            file_info = get_file_metadata(dataset=dataset, all=all, numevtsperfile = 0)
            metadata.update(file_info)

        return metadata
